import es
import playerlib
import weaponlib
import vecmath
import cmdlib
import effectlib
import gamethread

from path import path as Path
from math import degrees
import random
import time

import spe

class ESTManager(object):
    """
    This object is here to store all ESTCommand objects. It acts as a customized
    container allowing us to access the individual stored objects by a given
    name. This will be the manager for all the EST commands themselves and
    provide a global interface.
    """
    def __init__(self):
        """
        Default constructor, initalize variables.
        """
        self.commands = {}
        
    def __del__(self):
        """
        Default destructor, clear all the commands in turn executing the 
        desctructor of that object which unregisters that command object.
        """
        self.commands.clear()
        
    def __getitem__(self, name):
        """
        Returns the EST command object specified by the name.
        
        @param str name The name of the EST command
        @return ESTCommand The object designated by the name
        """
        name = name.lower()
        if name.startswith("est_"):
            name = name[4:]
        return self.commands[name]
    
    def __delitem__(self, name):
        """
        Removes the EST command object which will unregister it.
        
        @param str name The name of the EST command
        """
        name = name.lower()
        if name.startswith("est_"):
            name = name[4:]
        self.commands.__delitem__(name)
    
    def __contains__(self, name):
        """
        Test to see if the EST command is currently in use
        
        @param str name The name of the EST command
        @return bool Whether or not the EST command exists
        """
        name = name.lower()
        if name.startswith("est_"):
            name = name[4:]
        return name in self.commands
        
    def registerESTCommand(self, name, description, callback, argCount):
        """
        Registers a new EST command by the given parameters. This will register
        it with cmdlib.
        
        @param str name The name of the EST command
        @param str description The help message of the command; contains paramter list
        @param function callback The callback executed when the command is used
        @param int|tuple argCount The amount of arguments which are valid for the command
        """
        self.commands[name.lower()] = ESTCommand(name, description, callback, argCount)
    
    def unregisterESTCommand(self, name):
        """
        Unregisters an EST command.
        
        @param str name The name of the EST command
        """
        name = name.lower()
        if name.startswith("est_"):
            name = name[4:]
        del self.commands[name]
        
    def dumpCommands(self):
        """
        Dumps the entire command list to 
        """
        es.dbgmsg(0, "\n\nEST Command list:")
        es.dbgmsg(0, "-" * 50)
        for commandName in sorted(self.commands):
            commandObj = self.commands[commandName]
            commandObj.displayHelp()
        es.dbgmsg(0, "-" * 50)
        es.dbgmsg(0, "\n\n")
            
    
class ESTCommand(object):
    """
    This is an individual command object. This object will be created for all
    EST commands. This wraps the cmdlib object in a specific API which is very
    EST related. This enables us to have consistency throughout all of our 
    objects as it is handled automatically.
    """
    def __init__(self, name, description, callback, argCount):
        """
        Default constructor, intialise all variables and register the command.
        We need to check if the command is lowercase, if not, register 2 commands
        (one for the lowercase alternative as well.)
        
        @param str name The name of the EST command
        @param str description The help message of the command; contains paramter list
        @param function callback The callback executed when the command is used
        @param int|tuple argCount The amount of arguments which are valid for the command
        """
        if not name.startswith("est_"):
            name = "est_%s" % name
        self.callback = callback
        self.name = name
        if not self.name.islower():
            self.lowerName = name.lower()
        self.description = description
        if not hasattr(argCount, "__iter__"):
            argCount = (argCount, )
        self.argCount = argCount
        cmdlib.registerServerCommand(name, self._callback, description)
        cmdlib.registerServerCommand(self.lowerName, self._callback, description, True)
    
    def __del__(self):
        """
        Default destructor. Unregister the server commands.
        """
        cmdlib.unregisterServerCommand(name)
        if not self.name.islower():
            cmdlib.unregisterServerCommand(self.name.lower())
        
    def displayHelp(self):
        """
        Display a help message about how to use this command. This will also
        display the expected arg count.
        """
        if len(self.argCount) == 1:
            argCount = self.argCount[0]
        else:
            argCount = self.argCount
        es.dbgmsg(0, "%s %s : expects %s args" % (self.name, self.description, argCount) )
        
    def _callback(self, args):
        """
        Executed when this command is executed. Test the argument length, if we
        have any conflicts, display an error message and return early. Otherwise
        we need to follow to the actuall callback function.
        
        @param args A list of arguments after the command
        """
        if len(args) not in self.argCount:
            es.dbgmsg(0, "EST ERROR: Incorrect argument usaged, the correct usage is - %s %s" % (self.name, self.description) )
            return
        if self.callback(args) is False:
            es.dbgmsg(0, "EST ERROR: An error occursed in the line: %s %s" % (self.name, " ".join(args) ) )
            
class DelayManager(object):
    """
    This object manages all of the delay objects for individual players. This
    provides an easy to access interface to store and manage all Delay objects
    based on a key (the player's userid.)
    """
    def __init__(self):
        """
        Default constructor, initialise all variables.
        """
        self.delayObjects = {}
    
    def __getitem__(self, userid): 
        """
        Executed when the dict referencing command is used on this object.
        Assume we want to get the DelayObject value from the userid passed.
        
        @param int userid The ID of the user that we want to get the DelayObject
        @return DelayObject|None The releveant DelayObject for the userid
        """
        userid = int(userid)
        if userid in self.delayObjects:
            return self.delayObjects[userid]
        return None
    
    def __contains__(self, userid):
        """
        Executed when we test for an object's validation within this object.
        Assume we want to test if a player has been added.
        
        @param int userid The ID of the user that we want to get.
        """
        userid = int(userid)
        return bool(userid in self.delayObjects)
        
    def __delitem__(self, userid):
        """
        Executed when the del command is executed on this object with an
        attribute. Assume we want to remove the player.
        
        @param int userid The ID of the user that we want to get.
        """
        userid = int(userid)
        if userid in self.delayObjects:
            del self.delayObjects[userid]
            
    def addPlayer(self, userid):
        """
        Adds a player into the delay management.
        
        @param int userid The ID of the user that we want to get.
        """
        userid = int(userid)
        if userid not in self.delayObjects:
            self.delayObjects[userid] = DelayObject(userid)
        return self.delayObjects[userid]
            
    def removePlayer(self, userid):
        """
        Removes a player from the delay management.
        
        @param int userid The ID of the user that we want to get.
        """
        self.__delitem__(userid)
        
class DelayObject(object):
    """
    This delay object manages all gamethread delays for a player. This allows
    us to keep track of all the delays for a player and cancel them all on
    player_death or round_end. This also means when a player leaves the server
    their delays will automatically be cancelled so no errors.
    """
    def __init__(self, userid):
        """
        Default constructor, intialise all the variables.
        
        @param int userid The ID of the user that this object represents.
        """
        self.userid = int(userid)
        self.delays = set()
        
    def __del__(self):
        """
        Default destructor, executes when this object is destroyed. Remove all
        of the delays that this player has active.
        """
        self.removeAllDelays()
    
    def removeAllDelays(self):
        """
        Remove any outstanding delays the user has active.
        """
        for delay in self.delays:
            gamethread.cancelDelayed(delay)
        self.delays.clear()
    
    def addDelay(self, time, name, function, *args):
        """
        Short wrapper to delay a function
        
        @param float time The time to delay the function
        @param str name The name of this delay (will be made unique)
        @param func function The function address of the function to be executed
        @param tuple args Any additional arguments to pass to the function
        """
        name = "est_%s_user_%s" % (name, self.userid) 
        gamethread.delayedname(time, name, function, args)
        self.delays.add(name)
        
    def cancelDelay(self, name):
        """
        Cancel the current outstanding delay referenced by the name.
        
        @param str name The name of the delay to cancel. 
        """
        name = "est_%s_user_%s" % (name, self.userid)
        gamethread.cancelDelayed(name)
        if name in self.delays:
            self.delays.remove(name)
            
class FileManager(object):
    def __init__(self):
        """
        Default constructor, ensure that the default objects are initialised.
        This will register all the unallowed files to be accessed by file modes.
        This will also give us allowed file extensions.
        """
        self.disallowedFiles = {}
        self.disallowedFiles["gameinfo.txt"] = ("w",)
        self.disallowedFiles["motd.txt"] = ("w",)
        self.disallowedFiles["cfg/server.cfg"] = ("w", "r")
        self.disallowedFiles["cfg/autoexec.cfg"] = ("w",)
        self.disallowedFiles["cfg/banned_user.cfg"] = ("w",)
        self.disallowedFiles["cfg/banned_ip.cfg"] = ("w",)
        self.disallowedFiles["cfg/mani_server.cfg"] = ("w",)
        self.disallowedFiles["cfg/mani_admin_plugin/clients.txt"] = ("w",)
        self.allowedExtensions = (".info", ".cfg", ".txt", ".ini")
        self.openFileType = {'r':'r', 'w':'a+'}
        self.files = {}
        self.currentFileHandle = -1
        
    def __del__(self):
        """
        Default destructor, ensure that all opened path files are closed
        automatically so the streams aren't left open if this is unloaded for
        example.
        """
        for fileObject in self.files.itervalues():
            fileObject.close()
        self.files.clear()
        
    def openFile(self, args):
        """
        Open a file stream object and store the open stream in a dictionary and
        return the handle id to that object. Parameter 2 "fileType" is optional,
        if not specified assume 'r' mode.
        
        @param str args[0] var The variable to assign the file handle
        @param str args[1] fileType The method of opening the file either 'w' or 'r'
        @param str args[2] filePath A relative filePath to the file to open
        """
        var = args[0]
        fileType = 'r'
        if len(args) == 3:
            filePath = args[2].replace("\\", "/").strip().replace("//", "/")
            fileType = args[1].lower()
        else:
            filePath = args[1].replace("\\", "/").strip().replace("//", "/")            
        if ".." in filePath:
            es.dbgmsg(0, "EST ERROR: You are not allowed previous directories in file paths")
            es.ServerVar(var).set(-1)
            return False
        if filePath[1] == ":":
            es.dbgmsg(0, "EST ERROR: You cannot specify direct paths, you must use relative paths")
            es.ServerVar(var).set(-1)
            return False
        if fileType not in ("r", "w"):
            es.dbgmsg(0, "EST ERROR: Filetype %s is not a valid file type, only 'r' and 'w' are acceptable"  % fileType)
            es.ServerVar(var).set(-1)
            return False
        if filePath in self.disallowedFiles and fileType in self.disallowedFiles[filePath]:
            es.dbgmsg(0, "EST ERROR: You are not allowed to open file %s in stream '%s'" % (filePath, fileType))
            es.ServerVar(var).set(-1)
            return False
        if not filePath.endswith(self.allowedExtensions):
            es.dbgmsg(0, "EST ERROR: Your path does not end with a valid file type")
            es.ServerVar(var).set(-1)
            return False
        realPathFile = Path(str(es.ServerVar("eventscripts_gamedir"))).joinpath(args[2])
        if realPathFile.isfile():
            self.currentFileHandle += 1
            self.files[self.currentFileHandle] = realPathFile.open(self.openFileType[fileType])
            es.ServerVar(var).set(self.currentFileHandle)
        else:
            es.dbgmsg(0, "EST ERROR: File %s does not exist" % self.filePath)
            es.ServerVar(var).set(-1)
            return False
            
    def closeFile(self, args):
        """
        Closes a current open file stream.
        
        @param args[0] int fileHandle The handle to the file to close
        """
        fileHandle = str(args[0])
        if not fileHandle.isdigit():
            es.dbgmsg(0, "EST ERROR: Expected a file handle value for est_closefile, instead received %s" % fileHandle)
            return False
        fileHandle = int(fileHandle)
        if fileHandle in self.files:
            self.files[fileHandle].close()
            del self.files[fileHandle]
        else:
            es.dbgmsg(0, "EST ERROR: File handle %s is not currently open" % fileHandle)
            return False
            
    def countLines(self, args):
        """
        Count the number of lines in a current file.,
        
        @param str args[0] variable The variable to hold 
        @param int args[1] handle The file handle to retrieve the lines for
        """
        fileHandle = str(args[1])
        if not fileHandle.isdigit():
            es.dbgmsg(0, "EST ERROR: Expected a file handle value for est_countlines, instead received %s" % fileHandle)
            return False
        fileHandle = int(fileHandle)
        if fileHandle in self.files:
            count = self.files[fileHandle].read().count("\n") + 1
            self.files[fileHandle].seek(0)
            es.ServerVar(args[0]).set(count)
        else:
            es.dbgmsg(0, "EST ERROR: File handle %s is not currently open" % fileHandle)
            return False
            
    def openFiles(self, args):
        """
        Prints out all current open file streams used by EST.
        """
        es.dbgmsg(0, "Files In memory by ES_Tools...")
        es.dbgmsg(0, "------------------------------")
        es.dbgmsg(0, "Files Open by the engine...")
        for handle, fileObject in self.files.items():
            es.dbgmsg(0, "Handle %s - %s (mode %s)" % (handle, fileObject.name, fileObject.mode) )
        es.dbgmsg(0, "------------------------------")  
        
    def printOpenFile(self, args):
        """
        Prints out all text from a current open file
        
        @param args[0] int fileHandle A handle to a file.
        """  
        fileHandle = str(args[0])
        if not fileHandle.isdigit():
            es.dbgmsg(0, "EST ERROR: Expected a file handle value for est_printopenfile, instead received %s" % fileHandle)
            return False
        fileHandle = int(fileHandle)
        if fileHandle in self.files:
            es.dbgmsg(0, "------------------------------")
            for line in self.files[fileHandle].read().split("\n"):
                es.dbgmsg(0, line.strip("\r"))
            self.files[fileHandle].seek(0)
            es.dbgmsg(0, "------------------------------")
        else:
            es.dbgmsg(0, "EST ERROR: File handle %s is not currently open" % fileHandle)
            return False
            
    def fileExists(self, args):
        """
        Retrieves whether or not a relative file exists or not.
        
        @param args[0] str Var The variable to assign the boolean value
        @param args[1] str Path The relative path to test for existance
        """
        realPathFile = Path(str(es.ServerVar("eventscripts_gamedir"))).joinpath(args[1])
        es.ServerVar(args[0]).set({True:1, False:0}[realPathFile.isfile()])
        
    def getLine(self, args):
        """
        Returns the line by the given line number.
        
        @param args[0] str variable The variable to assign the current line to
        @param args[1] int fileHandle The Id of the filePath which is the open file
        @param args[2] int lineNumber The number of the line wanted
        """
        fileHandle = str(args[1])
        if not fileHandle.isdigit():
            es.dbgmsg(0, "EST ERROR: Expected a file handle value for est_getline, instead received %s" % fileHandle)
            return False
        fileHandle = int(fileHandle)
        if fileHandle in self.files:
            lines = self.files[fileHandle].read().split("\n")
            if len(lines) < int(args[2]):
                es.dbgmsg(0, "EST ERROR: There are not %s lines in file %s" % (args[2], self.files[fileHandle].name))
                return False
            es.ServerVar(args[0]).set(lines[int(args[2]) - 1].strip("\r"))
            self.files[fileHandle].seek(0)
        else:
            es.dbgmsg(0, "EST ERROR: File handle %s is not currently open" % fileHandle)
            return False
            
    def deleteLine(self, args):
        """
        Removes a line by a given line id number.
        
        @param args[0] int fileHandle The Id of the filePath which is the open file
        @param args[1] int lineNumber The number of the line to remove
        """
        fileHandle = str(args[0])
        if not fileHandle.isdigit():
            es.dbgmsg(0, "EST ERROR: Expected a file handle value for est_deleteline, instead received %s" % fileHandle)
            return False
        fileHandle = int(fileHandle)
        if fileHandle in self.files:
            if self.files[fileHandle].mode == 'r':
                es.dbgmsg(0, "EST ERROR: Can't delete a line whilst in 'r' mode.")
                return False
            else:
                lines = self.files[fileHandle].read().split("\n") # Load lines into memory
                if len(lines) < int(args[1]):
                    es.dbgmsg(0, "EST ERROR: There are not %s lines in file %s" % (args[2], self.files[fileHandle].name))
                    return False
                del lines[int(args[1]) - 1] # Delete the line from memory
                self.files[fileHandle].truncate(0) # Destroy all text
                self.files[fileHandle].write("\n".join(lines)) # Write all lines back in
        else:
            es.dbgmsg(0, "EST ERROR: File handle %s is not currently open" % fileHandle)
            return False
    
    def insertLine(self, args):
        """
        Inserts a line before the given line id number
        
        @param args[0] int fileHandle The ID of the filePath which is the open file
        @param args[1] int lineNumber The number of the line to insert before (-1 = at the end)
        @param args[2] str line The text to insert before.
        """
        fileHandle = str(args[0])
        if not fileHandle.isdigit():
            es.dbgmsg(0, "EST ERROR: Expected a file handle value for est_insertline, instead received %s" % fileHandle)
            return False
        fileHandle = int(fileHandle)
        if fileHandle in self.files:
            if self.files[fileHandle].mode == 'r':
                es.dbgmsg(0, "EST ERROR: Can't insert a line whilst in 'r' mode.")
                return False
            else:
                lines = self.files[fileHandle].read().split("\n")
                if len(lines) < int(args[1]) and int(args[1]) != -1:
                    es.dbgmsg(0, "EST ERROR: There are not %s lines in file %s" % (args[2], self.files[fileHandle].name))
                    return False
                if int(args[1]) == -1:
                    lines.append(args[3])
                else:
                    lines.insert(int(args[1]) - 1, args[3])
                self.files[fileHandle].truncate(0)
                self.files[fileHandle].write("\n".join(lines))
        else:
            es.dbgmsg(0, "EST ERROR: File handle %s is not currently open" % fileHandle)
            return False
    
class GravityManager(object):
    """ Class to manager the tick listener, and to manage the players gravity """
    def __init__(self):
        """ Default constructor; assign all local methos here """
        self.gravityList = {}
       
    def __del__(self):
        """
        Default destructor, executed when this object is destored. Cancel
        the delay name.
        """
        gamethread.cancelDelayed("gravity_check")
       
    def __getitem__(self, userid):
        """
        Executed automatically when the singleton is indexed, return the
        gravity object
        
        @param userid - userid to return to GravityObject() of
        @return GravityObject - that is relevant to the player
        """
        userid = int(userid)
        if self.__contains__(userid):
            return self.gravityList[userid]
        return None
    
    def __delitem__(self, userid):
        """
        Executed automatically when del instnace[key] is executed. Remove
        the player
        
        @param userid - userid to delete
        """
        self.removePlayer(userid)
        
    def __contains__(self, userid):
        """
        Executed automatically when we run a test to see if a userid exists
        within the singleton.
        
        @param userid - the id of the user you want to test for validity
        @return boolean - whether or not the player exists
        """
        userid = int(userid)
        return bool(userid in self.gravityList)
       
    def addPlayer(self, userid, amount):
        """ 
        Check if there are already any players in the gravityChange list.
        If there isn't, start the tick listener. Following this, check
        if the userid is in the dictionary, if so, remove them. Then create
        a new instance.
        
        @param userid - the user to set the gravity to
        @param amount - the new amount of gravity to assign
        """
        userid = int(userid)
        
        if not self.gravityList:
            gamethread.delayedname(0.25, 'gravity_check', self._ticker)
            
        if self.__contains__(userid):
            self.removePlayer(userid)
            
        self.gravityList[userid] = GravityObject(userid, float(amount) )
        self.gravityList[userid].update()
        self.gravityList[userid].reset()
       
    def removePlayer(self, userid):
        """ 
        Check if the player is in the dictioanry. If so, reset their gravity to 1
        and delete their instance from the dictionary. If there are no more players
        within the gravityList, remove the tick listener
        
        @param userid - the use who to check
        """
        userid = int(userid)
        
        if self.__contains__(userid):
            del self.gravityList[userid]
            
        if not self.gravityList:
            gamethread.cancelDelayed('gravity_check')
       
    def clearList(self):
        """
        Loop through all the players, reset their gravity to 1, delete the gravity
        list then unregister the tick listener.
        """
        self.gravityList.clear()
        gamethread.cancelDelayed('gravity_check')
       
    def _ticker(self):
        """
        Here we loop through all of the players, and check their gravity etc.
        """
        for player in self.gravityList.itervalues():
            player.update()
        gamethread.delayedname(0.25, 'gravity_check', self._ticker)        
        
class GravityObject(object):
    """
    This class managers all the players, their current values and any functions
    reloving around the player.
    """
    def __init__(self, userid, amount):
        """
        Default constructor. Assign default values
        
        @param userid - the user who this gravity object represents
        @param amount - the default amount of gravity to assign
        """
        self.userid = int(userid)
        self.amount = amount
        
    def __del__(self):
        """ Default deconstructor; ensure their gravity is assigned back to 0 """
        self.reset(1)
        
    def update(self):
        """
        This method updates a player's settings. If they've changed, they will
        reset the players gravity
        """
        if es.exists('userid', self.userid):
            self.reset()
        
    def reset(self, amountOverride = None):
        """
        This method resets a player's gravity either back to the default amount
        or if overridden, then to that amount.
        
        @param OPTIONAL amountOverride - if this is set to an integral amount,
                                         that value will take presedence over the
                                         default value.
        """
        if amountOverride is None:
            es.server.es_fire(self.userid, "!self", "addoutput", '"gravity %s"' % self.amount, "0.1", "1")
        else:
            es.server.es_fire(self.userid, "!self", "addoutput", '"gravity %s"' % amountOverride, "0.1", "1")
            
class RestrictionManager(object):
    """
    This object manages all of the restrictions for individual players. This
    provides an easy to access interface to store and manage all restriction objects
    based on a key (the player's userid.)
    """
    def __init__(self):
        """
        Default constructor, initialise all variables.
        """
        self.restrictions = {}
    
    def __getitem__(self, userid): 
        """
        Executed when the dict referencing command is used on this object.
        Assume we want to get the RestrictObject value from the userid passed.
        
        @param int userid The ID of the user that we want to get the RestrictObject
        @return RestrictObject|None The releveant RestrictObject for the userid
        """
        userid = int(userid)
        if userid in self.restrictions:
            return self.restrictions[userid]
        return None
    
    def __contains__(self, userid):
        """
        Executed when we test for an object's validation within this object.
        Assume we want to test if a player has been added.
        
        @param int userid The ID of the user that we want to get.
        """
        userid = int(userid)
        return bool(userid in self.restrictions)
        
    def __delitem__(self, userid):
        """
        Executed when the del command is executed on this object with an
        attribute. Assume we want to remove the player.
        
        @param int userid The ID of the user that we want to get.
        """
        userid = int(userid)
        if userid in self.restrictions:
            del self.restrictions[userid]
            
    def addPlayer(self, userid):
        """
        Adds a player into the restrict management.
        
        @param int userid The ID of the user that we want to get.
        """
        userid = int(userid)
        if userid not in self.restrictions:
            self.restrictions[userid] = RestrictionObject(userid)
        return self.restrictions[userid]
            
    def removePlayer(self, userid):
        """
        Removes a player from the restrict management.
        
        @param int userid The ID of the user that we want to get.
        """
        self.__delitem__(userid)
        
    def restrict(self, args):
        """
        Restricts a group of players weapons so they can't buy / pickup the
        ones restricted.
        
        @param str args[0] EST Player Filter
        @param str|int args[1] weapon tag The weapon name / slot to restrict
        """
        for player in getUseridsFromFilters(args[0]):
            restrictionObject = self[player]
            if restrictionObject is not None: 
                restrictionObject.restrict(args[1])
        
    def restrictAll(self, args):
        """
        Restricts all weapons (including knife and C4) from a player. This means
        that they won't be able to buy / pickup any weapons unless they're
        unrestricted after.
        
        @param str args[0] EST Player Filter
        """
        for player in getUseridsFromFilters(args[0]):
            for weapon in weaponlib.getWeaponList():
                restrictionObject = self[player]
                if restrictionObject is not None:
                    restrictionObject.restrict(weapon.name)
        
    def unrestrict(self, args):
        """
        unrestricts a group of players weapons so they can buy / pickup the
        ones unrestricted again.
        
        @param str args[0] EST Player Filter
        @param str|int args[1] weapon tag The weapon name / slot to restrict
        """
        for player in getUseridsFromFilters(args[0]):
            restrictionObject = self[player]
            restrictionObject.unrestrict(args[1])
    
    def unrestrictAll(self, args):
        """
        Unrestricts all restrictions for a filter group. This means that the
        players will be able to buy / pickup any weapon again.
        
        @param str args[0] EST Player Filter
        """
        for player in getUseridsFromFilters(args[0]):
            for weapon in weaponlib.getWeaponList():
                restrictionObject = self[player]
                restrictionObject.unrestrict(weapon.name)
                
    def showRestrictions(self, args):
        """
        Shows all restrictions for all the players given as the filter.
        
        @param str args[0] EST Player Filter
        """
        for player in getUseridsFromFilters(args[0]):
            self[player].showRestrictions()
        
class RestrictionObject(object):
    """
    This object is created for all players on a server. This allows us to
    store and handle all restrictions on a user level rather than a global
    level.
    """
    def __init__(self, userid, restrictions=[]):
        """
        Default constructor. Ensure that we register all the default storage
        containers.
        
        @param int userid The ID of the user
        @param iterable restrictions A list of default weapon restrictions
        """
        self.userid = int(userid)
        self.restrictions = set(restrictions)
    
    def restrict(self, weapon):
        """
        Restrict a weapon for this user. This will also check for any unhandled
        weapons as well.
        
        @param str weapon The name of the weapons to restrict. 
        """
        weapons = _getWeapon(weapon)
        for weapon in weapons:
            if weapon is None:
                return
            self.restrictions.add(weapon.name)
            self.handlePickup(weapon.name)
        
    def unrestrict(self, weapon):
        """
        Unrestricts a weapon for this user. This will weapon the ban from being
        able to purchase / pick up this weapon.
        
        @param str weapon The name of the weapons to restrict.
        """
        weapons = _getWeapon(weapon)
        for weapon in weapons:
            weapon = weaponlib.getWeapon(weapon)
            if weapon is None:
                return
            if weapon.name in self.restrictions:
                self.restrictions.remove(weapon.name)
    
    def handleClientCommandFilter(self, args):
        """
        This is called when a client command is executed on this person. If
        we encounter a buy command, check the weapon to see if it's restricted.
        If it is, the tell them, and return False. If everything's okay, return 
        True.
        
        @param list args[0] A list of arguments passed into the command filter
        """
        if args:
            if args[0].lower().strip() == "buy":
                if len(args) >= 1:
                    weapon = args[1].lower()
                    weapon = weaponlib.getWeapon(weapon)
                    if weapon is None:
                        es.tell(self.userid, '#green', 'The weapon %s is resticted' % weapon.name.split("_", 1)[1])
                        return False
                    if weapon.name in self.restrictions:
                        es.tell(self.userid, '#green', 'The weapon %s is resticted' % weapon.name.split("_", 1)[1])
                        return False
        return True
        
    def handlePickup(self, item):
        """
        This executed when a user picks up a weapon. Test to see if the weapon
        is restricted, and if so, remove it.
        
        @param str item The weapon / item which was just picked up.
        """
        weapon = weaponlib.getWeapon(item)
        if weapon is None:
            return
        if weapon.name in self.restrictions:
            index = spe.getWeaponIndex(self.userid, weapon.name)
            if index is not None:
                spe.removeEntityByIndex(index)
                es.cexec(self.userid, 'lastinv')
                es.tell(self.userid, '#green', 'The weapon %s is resticted' % weapon.name.split("_", 1)[1])
            
    def showRestrictions(self):
        """
        Shows all restrictions for this player:
        """
        es.dbgmsg(0, "\nRestictions for player %s" % es.getplayername(self.userid))
        for restriction in self.restrictions:
            es.dbgmsg(0, "Restricted: %s" % restriction) 
    
def clientCommandFilter(userid, args):
    """
    Executed before the server responds to a client command. This allows us to
    hook the drop command, and give them an undropable weapon which forces them
    to "drop" the item, then remove the item from their hands.
    
    @param int userid the ID of the user who executed the command
    @param tuple|mixed args An assortmnet of the arguments passed as the command
    @return bool Whether or not the server will process the client command 
    """
    if int(advancedDrop):
        if len(args) == 1:
            if args[0].lower().strip() == "drop":
                weapon = playerlib.getPlayer(userid).weapon.replace("weapon_", "") 
                if weapon in ("knife", "hegrenade", "flashbang", "smokegrenade"):
                    weapon = weaponlib.getWeapon(weapon)
                    handle = es.getplayerhandle(userid) 
                    #es.server.es_give(userid, weapon.name)
                    spe.giveNamedItem(userid, weapon.name)
                    index = getWeaponIndex(userid, weapon.name)
                    if index is not None:
                        spe.removeEntityByIndex(index)
                        es.cexec(userid, 'lastinv')
                    return False
            if userid in restrictions:
                returnVal = restrictions[userid].handleClientCommandFilter(args)
                if returnVal is False:
                    return False
    return True
    
def sayFilter(userid, text, teamOnly):
    """
    Executes before a player's message is displayed in chat. If dead chat is on
    then we need to ensure that we send any dead player's messages to the relevant
    living players.
    
    @param int userid The ID of the user who sent the message
    @param str text The message which was sent
    @param bool teamOnly Whether or not the message was sent in team only chat 
    @return tuple|str|int A 3 item tuple containing the return userid, text and team status of the message
    """
    if userid in muted:
        return (0, "", 0)
    lastMessages[userid] = text
    if int(deadChat):
        if es.getplayerprop(userid, "CBasePlayer.pl.deadflag"):
            name = es.getplayername(userid)
            team = es.getplayerteam(userid)
            for player in es.getUseridList():
                if not es.getplayerprop(player, "CBasePlayer.pl.deadflag"):
                    if (teamOnly and es.getplayerteam(player) == team) or not teamOnly:
                        es.tell(player, '#default', "*dead* %s: %s" % (name, text) )
    return (userid, text, teamOnly)
         
est          = ESTManager()
delays       = DelayManager()
files        = FileManager()
gravity      = GravityManager()
restrictions = RestrictionManager()
lastMessages = {}
entNames     = {}
muted        = set()
startTime    = None

advancedDrop = es.ServerVar("est_allowadvanceddrop", "Allow the drop of grenades / knives", 0)
deadChat     = es.ServerVar("est_dead_chat", "Allows dead players to talk in chat to living players", 0) 
est_version  = es.ServerVar("est_version", "The version of EventScripts", "0.420")

def load():
    """
    Executed when this addon loads, registers all the EST commands.
    """
    global startTime
    startTime = time.time()
    es.addons.registerClientCommandFilter(clientCommandFilter)
    es.flags('remove', 'cheat', 'endround')
    
    est.registerESTCommand("AddDownload", '<path>', addDownload, 1)
    est.registerESTCommand("CSay", '<players> <message>', cSay, 2)
    est.registerESTCommand("HSay", '<players> <message>', hSay, 2)
    est.registerESTCommand("CurTime", '<var>', getCurrentTime, 1)
    est.registerESTCommand("Damage", '<attacker> <victim> <amount of damage>', damage, 3)
    est.registerESTCommand("DeleteRagDolls", '[est filter]', removeRagdolls, (0, 1))
    est.registerESTCommand("Freeze", '<est filter> <1/0>', freeze, 2)
    est.registerESTCommand("DropWeapon", '<est filter>', drop, 2)
    est.registerESTCommand("EndRound", '', endRound, 0)
    est.registerESTCommand("Dissolve", '<classname/index>', dissolve, 1)
    est.registerESTCommand("Name", '<userid> <name>', changeName, 2)
    est.registerESTCommand("UpTime", '<var>', uptime, 1)
    est.registerESTCommand("StripPlayer", '<est filter>', stripPlayer, 1)
    est.registerESTCommand("Speed", '<est filter> <value>', speed, 2)
    est.registerESTCommand("Teleport", '<player> <x> <y> <z>', teleport, 4)
    est.registerESTCommand("Slay", '<est filter>', slay, 1)
    est.registerESTCommand("Drug", '<est filter> [time]', drug, (1, 2))
    est.registerESTCommand("Drunk", '<est filter> [time]', drunk, (1, 2))
    est.registerESTCommand("Give", '<est filter> <entity>', give, 2)
    est.registerESTCommand("God", '<est filter> <1/0>', god, 2)
    est.registerESTCommand("GetWallBetween", '<var> <x> <y> <z> <targetx> <targety> <targetz>', getWallBetween, 7)
    est.registerESTCommand("GetChat", '<var> <name/userid>', getChat, 2)
    est.registerESTCommand("Ban", '<est filter> <time> [reason]', ban, (2, 3))
    est.registerESTCommand("Kick", '<est filter> [reason]', kick, (1, 2))
    est.registerESTCommand("Burn", '<est filter> [time]', burn, (1, 2))
    est.registerESTCommand("Extinguish", '<est filter>', extinguish, 1)
    est.registerESTCommand("CExec", '<est filter> <command>', cexec, 2)
    est.registerESTCommand("Help", '[command]', getHelp, (0,1))
    est.registerESTCommand("Fade", '<est filter> <type> <fade time> <total time> <red> <green> <blue> <alpha>', fade, 8)
    est.registerESTCommand("JetPack", '<est filter> <1/0>', jetpack, 2)
    est.registerESTCommand("NoClip", '<est filter> <1/0>', noclip, 2)
    est.registerESTCommand("LightStyle", '<stylestring> <stylenum>', changeLightStyle, 2)
    est.registerESTCommand("MutePlayer", '<est filters>', mutePlayer, 1)
    est.registerESTCommand("UnMutePlayer", '<est filters>', unmutePlayer, 1)
    est.registerESTCommand("Team", '<est filters> <team index>', switchTeam, 2)
    est.registerESTCommand("Spawn", '<est filters> [force spawn]', spawn, (1, 2))
    est.registerESTCommand("Play", '<est filters> <sound>', playSoundForPlayers, 2)
    est.registerESTCommand("PlayEdict", '<entity index> <sound> [volume] [est filter]', emitSoundFromEntity, (2, 3, 4))
    est.registerESTCommand("PlayPlayer", '<userid> <sound> [volume] [est filter]', emitSoundFromPlayer, (2, 3, 4))
    est.registerESTCommand("PlayerCount", '<var> <est filter>', getPlayerCount, 2)
    est.registerESTCommand("PropCount", '<var>', getPropCount, 1)
    est.registerESTCommand("PropList", '', listAllProps, 0)
    est.registerESTCommand("PropShow", '[class type]',  showAllOffsets, (0, 1))
    est.registerESTCommand("RandPlayer", '<var> <est filter>', getRandomPlayer, 2)
    est.registerESTCommand("RemoveIdle", '[classname]', removeIdle, (0, 1))
    est.registerESTCommand("RemoveWeapon", '<est filter> <weapon name>', removeWeapon, 2)
    est.registerESTCommand("SetEntModel", '<ent index> <model path>', setEntModel, 2)
    est.registerESTCommand("SetEntityColor", '<ent index> <x> <y> <z> [alpha]', setEntColor, (4, 5))
    est.registerESTCommand("SetPlayerColor", '<est filters> <x> <y> <z> [alpha]', setPlayerColor, (4, 5))
    est.registerESTCommand("Shake", '<est filter> <time> <distance> <violence>', shake, 4)
    est.registerESTCommand("StopSound", '<est filter> <sound>', stopSound, 2)
    est.registerESTCommand("SetAim", '<est filter> <target userid/name> [add to z coord]', setAim, (2, 3))
    est.registerESTCommand("SetViewCoord", '<est filter> <x> <y> <z>', setViewCoord, 4)
    est.registerESTCommand("SetViewAngle", '<est filter> <pitch> <yaw> <roll>', setViewAngle, 4)
    est.registerESTCommand("ShowMenu", '<time> <est filter> <menu string>', showMenu, 3)
    est.registerESTCommand("SetAmmo", '<est filter> <weapon slot/name> <ammo>', setAmmo, 3)
    est.registerESTCommand("SetClipAmmo", '<est filter> <weapon slot/name> <clip ammo>', setClip, 3)
    est.registerESTCommand("PhysPush", '<est filter> <x force> <y force> <z force>', physPush, 4)
    est.registerESTCommand("PushTo", '<est filter> <x> <y> <z> <force>', pushTo, 5)
    est.registerESTCommand("Remove", '<ent index>', remove, 1)
    est.registerESTCommand("RoundDecimal", '<var> <number> <decimal digits>', roundDecimal, 3)
    est.registerESTCommand("Rocket", '<est filter>', rocket, 1)
    est.registerESTCommand("HasDefuser", '<var> <name/userid>', hasDefuser, 2)
    est.registerESTCommand("HasHelmet", '<var> <name/userid>', hasHelmet, 2)
    est.registerESTCommand("HasKevlar", '<var> <name/userid>', hasKevlar, 2)
    est.registerESTCommand("IsAlive", '<var> <name/userid>', isAlive, 2)
    est.registerESTCommand("IsBomber", '<var> <name/userid>', isBomber, 2)
    est.registerESTCommand("IsAlpha", '<var> <string>', isAlpha, 2)
    est.registerESTCommand("IsNumeric", '<var> <string>', isNumeric, 2)
    est.registerESTCommand("IsOnly", '<var> <string> <allowed chars>', isOnly, 3)
    est.registerESTCommand("ClassList", 'No arguments', classList, 0)
    est.registerESTCommand("GetEntClassName", '<var> <entity index>', getEntityClassName, 2)
    est.registerESTCommand("GetEntityIndex", '<var> <name/userid> <classname>', getEntityIndex, 3)
    est.registerESTCommand("GetEntName", '<var> <entity index>', getEntName, 2)
    est.registerESTCommand("SetEntName", '<entity index> <entity name>', setEntName, 2)
    est.registerESTCommand("GetAngleToPoint", '<userid/name> <x> <y> <z> <var pitch> <var yaw> <var roll>', getAngleToPoint, 7)
    est.registerESTCommand("GetViewAngle", '<userid/name> <var pitch> <var yaw> <var roll>', getViewAngle, 4)
    est.registerESTCommand("GetViewCoord", '<userid/name> <var x> <var y> <var z>', getViewCoord, 4)
    est.registerESTCommand("GetViewPlayer", '<userid/name> <var>', getViewPlayer, 2)
    est.registerESTCommand("GetViewProp", '<userid/name> <var>', getViewProp, 2)
    est.registerESTCommand("GetEyeCoord", '<userid/name> <x var> <y var> <z var>', getEyeCoord, 4)
    est.registerESTCommand("GetGunCoord", '<userid/name> <x var> <y var> <z var>', getGunCoord, 4)
    est.registerESTCommand("GetPlayerDistance", '<distance var> <azimuth var> <elevation var> <userid/name> <target id/target name>', getPlayerDistance, 5)
    est.registerESTCommand("GetIp", '<var> <userid/name>', getIP, 2)
    est.registerESTCommand("Getindex", '<var> <userid/name>', getIndex, 2)
    est.registerESTCommand("GetName", '<var> <userid/name>', getName, 2)
    est.registerESTCommand("GetUserid", '<var> <userid/name>', getUserid, 2)
    est.registerESTCommand("GetSteam", '<var> <userid/name>', getSteam, 2)
    est.registerESTCommand("GetTeam", '<var> <userid/name>', getTeam, 2)
    est.registerESTCommand("GetHE", '<var> <userid/name>', getHE, 2)
    est.registerESTCommand("GetFB", '<var> <userid/name>', getFB, 2)
    est.registerESTCommand("GetSG", '<var> <userid/name>', getSG, 2)
    est.registerESTCommand("GetPrimary", '<var> <userid/name>', getPrimary, 2)
    est.registerESTCommand("GetSecondary", '<var> <userid/name>', getSecondary, 2)
    est.registerESTCommand("KnifeOnly", '<var> <userid/name>', knifeOnly, 2)
    est.registerESTCommand("GetGun", '<var> <userid/name>', getGun, 2)
    est.registerESTCommand("MOTD_F", '<est filter> <title> <command> <file>', motdFromFile, 4)
    est.registerESTCommand("MOTD_W", '<est filter> <title> <command> <URL>', motdFromURL, 4)
    est.registerESTCommand("MOTD_T", '<est filter> <title> <command> <raw text>', motdFromRawText, 4)
    est.registerESTCommand("Msg", '<est filter> <message>', message, 2)
    est.registerESTCommand("Near", '<var> <est filter> <distance> <name/userid> <command>', near, 5)
    est.registerESTCommand("NearCoord", '<var> <est filter> <distance> <x> <y> <z> <command>', nearCoord, 7)
    est.registerESTCommand("MakeEntity", '<entity> <model> <x> <y> <z> [index var]', makeEntity, (5, 6))
    est.registerESTCommand("GetGravity", '<var> <userid/name>', getGravity, 2)
    est.registerESTCommand("SetGravity", '<est filter> <amount>', setGravity, 2)
    est.registerESTCommand("GetAmmo", '<var> <name/userid> <weapon slot / name>', getAmmo, 3)
    est.registerESTCommand("GetClipAmmo", '<var> <name/userid> <weapon slot / name>', getClip, 3)
    est.registerESTCommand("GetWeaponIndex", '<var> <name/userid> <weapon slot / name>', getWeaponIndex, 3)
    est.registerESTCommand("StrToUPPER", '<var> <string>', changeToUpper, 2)
    est.registerESTCommand("StrTolower", '<var> <string>', changeToLower, 2)
    est.registerESTCommand("StrLen", '<var> <string>', getStringLength, 2)
    est.registerESTCommand("TrimDecimal", '<var> <number> <decimal-position>', trimDecimal, 3)
    est.registerESTCommand("GetDeaths", '<var> <userid>', deathGet, 2)
    est.registerESTCommand("GetKills", '<var> <userid>', killGet, 2)
    est.registerESTCommand("Kills", '<est filter> <operator> <amount>', kills, 3)
    est.registerESTCommand("KillAdd", '<est filter> <amount>', killAdd, 2)
    est.registerESTCommand("KillSet", '<est filter> <amount>', killSet, 2)
    est.registerESTCommand("CvarAddFlag", '<cvar> <gamedll> <sponly> <notify> <cheat> <userinfo> <unlogged> <replicated>', addCVARFlags, 8)
    est.registerESTCommand("CvarDelFlag", '<cvar> <gamedll> <sponly> <notify> <cheat> <userinfo> <unlogged> <replicated>', deleteCVARFlags, 8)
    est.registerESTCommand("Armor", '<est filter> <operator> <amount>', armor, 3)
    est.registerESTCommand("ArmorAdd", '<est filter> <amount>', armorAdd, 2)
    est.registerESTCommand("SetArmor", '<est filter> <amount>', armorSet, 2)
    est.registerESTCommand("GetArmor", '<variable> <est player>', armorGet, 2)
    est.registerESTCommand("Health", '<est filter> <operator> <amount>', health, 3)
    est.registerESTCommand("HealthAdd", '<est filter> <amount>', healthAdd, 2)
    est.registerESTCommand("SetHealth", '<est filter> <amount>', healthSet, 2)
    est.registerESTCommand("GetHealth", '<variable> <est player>', healthGet, 2)
    est.registerESTCommand("Cash", '<est filter> <operator> <amount>', cash, 3)
    est.registerESTCommand("CashAdd", '<est filter> <amount>', cashAdd, 2)
    est.registerESTCommand("SetCash", '<est filter> <amount>', cashSet, 2)
    est.registerESTCommand("GetCash", '<variable> <est player>', cashGet, 2)
    est.registerESTCommand("ViewCone", '<var> <userid/name> <est filter> <cone angle> <cone distance> <command>', viewCone, 6)
    est.registerESTCommand("OpenFile", '<var> [file type (r / w)] <file path>', files.openFile, (2,3))
    est.registerESTCommand("OpenFiles", 'No arguments', files.openFiles, 0)
    est.registerESTCommand("PrintOpenFile", '<file handle>', files.printOpenFile, 1)
    est.registerESTCommand("CountLines", '<var> <file handle>', files.countLines, 2)
    est.registerESTCommand("CloseFile", '<file handle>', files.closeFile, 1)
    est.registerESTCommand("FileExists", '<var> <file path>', files.fileExists, 2)
    est.registerESTCommand("GetLine", '<var> <file handle> <line number>', files.getLine, 3)
    est.registerESTCommand("DeleteLine", '<file handle> <line number>', files.deleteLine, 2)
    est.registerESTCommand("InsertLine", '<file handle> <line number (-1 = append)> <text>', files.insertLine, 3)
    est.registerESTCommand("Restrict", '<est filter> <weapon name/slot>', restrictions.restrict, 2)
    est.registerESTCommand("UnRestrict", '<est filter> <weapon name/slot>', restrictions.unrestrict, 2)
    est.registerESTCommand("RestrictAll", '<est filter>', restrictions.restrictAll, 1)
    est.registerESTCommand("UnRestrictAll", '<est filter>', restrictions.restrictAll, 1)
    est.registerESTCommand("ShowRestrictions", '<est filter>', restrictions.showRestrictions, 1)
    
def unload():
    """
    Executed when this addon unloads, explicitly delete the est object from memory.
    """
    es.addons.unregisterClientCommandFilter(clientCommandFilter)
    
def round_start(event_var):
    """
    Executed when the round starts, clear the entity names list as all entities
    are refreshed each round.
    
    @param es.EventInfo event_var An automatically passed event instance. 
    """
    entNames.clear()
    
def player_disconnect(event_var):
    """
    Executed when a player disconnects, ensure that we clear up any instances / objets
    which may have been created by a player.
    
    @param es.EventInfo event_var An automatically passed event instance.
    """                                                                  
    userid = event_var['userid']
    if userid in muted:
        muted.remove(userid)
    del delays[userid]
    del restrictions[userid]
    
def player_activate(event_var):
    """
    Executed when a player activates on the server (joins). Initialise any
    direct player setting.
    
    @param es.EventInfo event_var An automatically passed event instance.
    """
    userid = event_var['userid']
    delays.addPlayer(userid)
    restrictions.addPlayer(userid)
    
def item_pickup(event_var):
    """
    Executed when a client picks up a weapon. If they're in the restrictions
    manager, the handle this request to ensure any resticted weapons are killed.
    
    @param es.EventInfo event_var An automatically passed event instance.
    """
    restriction = restrictions[event_var['userid']]
    if restriction is not None:
        restriction.handlePickup(event_var['item'])
    
def getPlayersFromFilters(filterList):
    """
    Return a list of players as playerlib filters from a given filter list. The
    filter list supports EST filter list selections (or playerlib selections)
    which accept either the filter tags (e.g. #c #t) or any valid user ident
    such as partial names, steamid's or user ids.
    
    @param string filterList A list of filters to search for the playerlib players
    @return tuple|playerlib.Player A list of players for the given filter list
    """
    playerTest = es.getuserid(filterList)
    if es.exists('userid', playerTest):
        return (playerlib.getPlayer(playerTest), )
    if not isinstance(filterList, basestring):
        return tuple()
    addPlayers = True
    players = set()
    estToPlayerlibFilters = {
    "c":"ct",
    "t":"t",
    "a":"all",
    "l":"alive",
    "s":"spec",
    "h":"human",
    "b":"bot",
    "d":"dead",
    "u":"un",
    "0":"un",
    "1":"spec",
    "2":"t",
    "3":"ct"
    }
    for filterLetter in filterList:
        if filterLetter == "#":
            addPlayers = True
        elif filterLetter == "!":
            addPlayers = False
        else:
            if filterLetter in estToPlayerlibFilters:
                playerlibFilterLetter = estToPlayerlibFilters[filterLetter]
            else:
                playerlibFilterLetter = filterLetter
            playerlibFilterLetter = "#" + playerlibFilterLetter
            playerList = set(playerlib.getPlayerList(playerlibFilterLetter))
            if addPlayers:
                players.update(playerList)
            else:
                players = players.difference(playerList)
    return players 

def getUseridsFromFilters(filterList):
    """
    Return a list of userids from a given filter list. The
    filter list supports EST filter list selections (or playerlib selections)
    which accept either the filter tags (e.g. #c #t) or any valid user ident
    such as partial names, steamid's or user ids.
    
    @param string filterList A list of filters to search for the playerlib players
    @return tuple|int A list of userids for the given filter list
    """
    return map(int, getPlayersFromFilters(filterList))
    
def applyOperatorToAttribute(player, attribute, operator, amount):
    """
    Applies an operator to a certain player. This will use the playerlib
    attributes so 'armor' will affect the player's armor etc. The only
    operators supported are +, - and = to increment, decrement and assign
    respectively as those are the operators supported by EST.
    
    @param playerlib.Player player The playerlib object of the user
    @param str attribute The playerlib attribute to apply the operator to
    @param str operator The operator type to apply
    @param int|float The amount to affect the value by.
    """
    if operator == "+":
        player[attribute] += amount
    elif operator == "-":
        player[attribute] -= amount
    elif operator == "=":
        player[attribute] = amount
    else:
        es.dbgmsg(0, "EST ERROR: est_%s cannot apply the operator %s. Only +, - and = are supported" % (attribute, operator))
    
def addDownload(args):
    """
    Registers a file for download
    
    @param str args[0] The path of the download file
    """
    es.stringtable("downloadbles", args[0])
    
def armor(args):
    """
    Applies an operator to a player's armor value.
    
    @param str args[0] EST Player filter
    @param str args[1] operator "+", "-" or "=" to reassign the armor
    @param int args[2] amount The amount to alter the armor by
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "armor", args[1], int(args[2]))
        
def armorAdd(args):
    """
    Adds a value to the player's current armor value
    
    @param str args[0] EST player filter
    @param int args[1] amount The amount to add to the armor
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "armor", "+", int(args[1]))
    
def armorSet(args):
    """
    Assigns a value to the player's armor value
    
    @param str args[0] EST player filter
    @param int args[1] amount The amount to assign to the armor
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "armor", "=", int(args[1]))
    
def armorGet(args):
    """
    Assigns a variable to hold the value of a player's current armor
    
    @param str args[0] variablename The name of the variable to use
    @param int args[1] EST Player An individual player to retrieve the armor
    """
    userid = es.getuserid(args[1])
    if es.exists('userid', userid):
        player = playerlib.getPlayer(userid)
        es.ServerVar(args[0]).set(int(player.armor))
        
def health(args):
    """
    Applies an operator to a player's health value.
    
    @param str args[0] EST Player filter
    @param str args[1] operator "+", "-" or "=" to reassign the health
    @param int args[2] amount The amount to alter the health by
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "health", args[1], int(args[2]))
        
def healthAdd(args):
    """
    Adds a value to the player's current health value
    
    @param str args[0] EST player filter
    @param int args[1] amount The amount to add to the health
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "health", "+", int(args[1]))
    
def healthSet(args):
    """
    Assigns a value to the player's health value
    
    @param str args[0] EST player filter
    @param int args[1] amount The amount to assign to the health
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "health", "=", int(args[1]))
    
def healthGet(args):
    """
    Assigns a variable to hold the value of a player's current health
    
    @param str args[0] variablename The name of the variable to use
    @param int args[1] EST Player An individual player to retrieve the health
    """
    userid = es.getuserid(args[1])
    if es.exists('userid', userid):
        player = playerlib.getPlayer(userid)
        es.ServerVar(args[0]).set(int(player.health))
        
def cash(args):
    """
    Applies an operator to a player's cash value.
    
    @param str args[0] EST Player filter
    @param str args[1] operator "+", "-" or "=" to reassign the cash
    @param int args[2] amount The amount to alter the cash by
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "cash", args[1], int(args[2]))
        
def cashAdd(args):
    """
    Adds a value to the player's current cash value
    
    @param str args[0] EST player filter
    @param int args[1] amount The amount to add to the cash
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "cash", "+", int(args[1]))
    
def cashSet(args):
    """
    Assigns a value to the player's cash value
    
    @param str args[0] EST player filter
    @param int args[1] amount The amount to assign to the cash
    """
    for player in getPlayersFromFilter(args[0]):
        applyOperatorToAttribute(player, "cash", "=", int(args[1]))
    
def cashGet(args):
    """
    Assigns a variable to hold the value of a player's current armor
    
    @param str args[0] variablename The name of the variable to use
    @param int args[1] EST Player An individual player to retrieve the cash
    """
    userid = es.getuserid(args[1])
    if es.exists('userid', userid):
        player = playerlib.getPlayer(userid)
        es.ServerVar(args[0]).set(int(player.cash))
        
def ban(args):
    """
    Bans a list of players via their IP and ID.
    
    @param str args[0] EST player filter
    @param int args[1] time Amount of time to ban for (seconds, 0=permanent)
    @param optional str args[2] reason The reason they were banned
    """
    for player in getPlayersFromFilter(args[0]):
        userid = int(player)
        address = int(player.address)
        reason = "Banned by admin"
        if len(args) > 2:
            reason = args[2]
        es.server.kickid(userid, reason)
        es.server.banid(args[1], player.steamid)
        es.server.addip(args[1], address)
    es.server.writeid()
    es.server.writeip()
    
def kick(args):
    """
    Kicks a player from the server with an optional reason
    
    @param str args[0] EST Player filter
    @param str args[1] reason Optional reason
    """
    for player in getUseridsFromFilter(args[0]):
        reason = "Kicked by admin"
        if len(args) > 1:   
            reason = args[1]
        es.server.kickid(userid, reason)
    
def burn(args):
    """
    Ignites an est player list
    
    @param str args[0] EST player filter
    @param optional float args[1] time Time to burn
    """
    for player in getPlayersFromFilter(args[0]):
        player.burn()
        if len(args) > 1 and float(args[1]):
            gamethread.delayed(float(args[1]), extinguish, (args[0]))
            
def extinguish(args):
    """
    Extinguishes players on fire
    
    @param str args[0] EST player filter
    """
    for player in getPlayersFromFilter(args[0]):
        player.extinguish()
        
def cexec(args):
    """
    Executes a client command on a player filter
    
    @param str args[0] EST player filter
    @param str args[1] command The command to execute on the players
    """
    command = '"%s"' % args[1]
    for player in getUseridsFromFilter(args[0]):
        es.server.es_cexec(player, commands)
        
def classList(args):
    """
    Dumps a list of server classes to the console.
    """
    es.server.es_dumpserverclasses()
    
def countEnts(args):
    """
    Count all the entities of a specific classname and returns it as a variable.
    
    @param str args[0] variable The variable to save the entity count to
    @param str args[1] classname The class name to get the amount on the server.
    """
    es.ServerVar(args[0]).set(len(es.createentitylist(args[1])))

def cSay(args):
    """
    Show a centre message to a specific person.
    
    @params args[0] str filters The list of filters to send the message to
    @params args[1] str message The message to send to the player's screen 
    """
    for player in getUseridsFromFilter(args[0]):
        es.centertell(player, args[1])
        
def hSay(args):
    """
    Show a hud message to a specific person.
    
    @params args[0] str filters The list of filters to send the message to
    @params args[1] str message The message to send to the player's screen 
    """
    usermsg.hudhint(getUseridsFromFilter(args[0]), args[1])
        
def getCurrentTime(args):
    """
    Retrieves the amount of seconds since epoch and stores it in a variable
    
    @param args[0] str variable The variable to store the time
    """
    es.ServerVar(args[0]).set(int(time.time()))
    
def addCVARFlags(args):
    """
    Add flags to a cvar. For a full list of the flags and descriptions see:
    http://developer.valvesoftware.com/wiki/Developer_Console_Control#The_FCVAR_flags
    
    @param args[0] str cvar The name of the CVAR to alter
    @param args[1] int gamedll flag Whether or not the FCVAR_GAMEDLL flag will be added
    @param args[2] int sponly flag Whether or not the FCVAR_SPONLY flag will be added (doesn't work)
    @param args[3] int notify flag Whether or not the FCVAR_NOTIFY flag will be added
    @param args[4] int cheat flag Whether or not the FCVAR_CHEAT flag will be added
    @param args[5] int userinfo flag Whether or not the FCVAR_USERINFO flag will be added (doesn't work)
    @param args[6] int unlogged flag Whether or not the FCVAR_UNLOGGED flag will be added
    @param args[7] int replicated flag Whether or not the FCVAR_REPLICATED flag will be added
    """
    _alterCVARFlags(args, 'add')
    
def deleteCVARFlags(args):
    """
    Removes flags from a cvar. For a full list of the flags and descriptions see:
    http://developer.valvesoftware.com/wiki/Developer_Console_Control#The_FCVAR_flags
    
    @param args[0] str cvar The name of the CVAR to alter
    @param args[1] int gamedll flag Whether or not the FCVAR_GAMEDLL flag will be removed
    @param args[2] int sponly flag Whether or not the FCVAR_SPONLY flag will be removed (doesn't work)
    @param args[3] int notify flag Whether or not the FCVAR_NOTIFY flag will be removed
    @param args[4] int cheat flag Whether or not the FCVAR_CHEAT flag will be removed
    @param args[5] int userinfo flag Whether or not the FCVAR_USERINFO flag will be removed (doesn't work)
    @param args[6] int unlogged flag Whether or not the FCVAR_UNLOGGED flag will be removed
    @param args[7] int replicated flag Whether or not the FCVAR_REPLICATED flag will be removed
    """
    _alterCVARFlags(args, 'remove')
    
def _alterCVARFlags(args, addOrRemove):
    """
    Alters CVAR vlaues with the given flags. Either adds or removes the flags
     
    @param tuple args List of arguments in the valid format (see params for deleteCVARFlags / addCVARFlags)
    @param str addOrRemove Either 'add' or 'remove' to decide how to alter the cvar flags 
    """
    cvar = args[0]
    gamedll = args[1]
    notify  = args[3]
    cheat   = args[4]
    unlogged = args[6]
    replicated = args[7]
    es.flags(addOrRemove, 'gamedll', gamedll, cvar)
    es.flags(addOrRemove, 'notify', notify, cvar)
    es.flags(addOrRemove, 'cheat', cheat, cvar)
    es.flags(addOrRemove, 'unlogged', unlogged, cvar)
    es.flags(addOrRemove, 'replicated', replicated, cvar)
    
def damage(args):
    """
    Damages a player and credits an attacker for the damage.
    
    @param args[0] int attacker The person to credit for the damage
    @param args[1] int victim The person who is delt the damage
    @param args[2] int damage The amount of damage to give
    """
    attacker = es.getuserid(args[0])
    victim = es.getuserid(args[1])
    if not attacker or not victim:
        es.dbgmsg(0, "EST ERROR: Cannot damage the player as either victim or attacker are invalid values")
        return False
    es.server.damage(victim, args[2], 32, attacker)
    
def deathGet(args):
    """
    Returns the deaths of a player and stores the values in a variable
    
    @param args[0] str var The name of the variable to store the death count in
    @param args[1] int userid The ID of the user to get the death count
    """
    userid == es.getuserid(args[1])
    if es.exists('userid', userid):
        player = playerlib.getPlayer(userid)
        es.ServerVar(args[0]).set(player.deaths)
    else:
        es.dbgmsg("EST ERROR: Player %s is not on the server" % args[1])
        es.ServerVar(args[0]).set(-1)
        
def killGet(args):
    """
    Returns the current kills (score) of the player.
    
    @param args[0] str var The name of the variable to store the kill count
    @param args[1] int userid The ID of the user to get the kill count
    """
    userid == es.getuserid(args[1])
    if es.exists('userid', userid):
        player = playerlib.getPlayer(userid)
        es.ServerVar(args[0]).set(player.deaths)
    else:
        es.dbgmsg("EST ERROR: Player %s is not on the server" % args[1])
        es.ServerVar(args[0]).set(-1)
        
def killSet(args):
    """
    Assigns a new value to the player's score board.
    
    @param args[0] str filter An EST player filter to change the score for
    @param args[1] int value The new value of the player's score
    """
    for player in getUseridsFromFilters(args[0]):
        es.server.score("set", player, args[1])
    
def killAdd(args):
    """
    Increments player's score boards score value by a given number.
    
    @param args[0] str filter An EST player filter to change the score for
    @param args[1] int value The value to add to the player's current value
    """
    for player in getUseridsFromFilters(args[0]):
        es.server.score("add", player, args[1])
    
def kills(args):
    """
    Applies an operator to a player's kill value.
    
    @param args[0] str filter An EST player filter to change the score for
    @param args[1] str operator Either + - or = to designate the action to apply to the current value
    @param args[2] int value The value to alter the current number by
    """ 
    for player in getPlayersFromFilters(args[0]):
        value = int(player.kills)
        if args[1] == "=":
            value = int(args[2])
        elif args[1] == "+":
            value += int(args[2])
        elif args[1] == "-":
            value -= int(args[2])
        else:
            es.dbgmsg(0, "EST ERROR: est_kills cannot apply the operator %s. Only +, - and = are supported" % operator)
            return False
        es.server.score("set", int(player), value)
        
def removeRagdolls(args):
    """
    Removes the ragdolls, if an argument is present applies it to a user list.
    
    @param optional args[0] str EST player filter
    """
    if args:
        for player in getUseridsFromFilters(args[0]):
            handle = es.getplayerhandle(player)
            for index in es.createentitylist("cs_ragdoll"):
                if es.getindexprop(index, "CBaseEntity.m_hOwnerEntity") == handle:
                    es.server.remove(index)
                    break
    else:
        es.server.remove("cs_ragdoll")
        
def freeze(args):
    """
    Freezes an EST player filter selection
    
    @param args[0] str EST Player filter
    @param args[1] int Whether this will freeze (1) or unfreeze (0)
    """
    for player in getPlayersFromFilters(args[0]):
        player.freeze = bool(int(args[1]))
        
def drop(args):
    """
    Forces an EST player selection to drop their weapons
    
    @param args[0] str EST Player Filters
    """
    for userid in getUseridsFromFilters(args[0]):
        es.sexec(userid, 'drop')
        
def endRound(args):
    """
    Ends the current round and plays a draw sound.
    
    Stolen from GODJonez's dmendround script:
        http://addons.eventscripts.com/addons/view/dmendround
    """
    for userid in es.getUseridList():
        es.playsound(userid, "radio/rounddraw.wav", 1.0)
    es.centermsg('Round Draw!')
    es.delayed(2, 'endround')
    
def dissolve(args):
    """
    Dissolve an entity index / class name with a cool special effect
    
    @param str|int args[0] ClassName / Index The target of the entity dissolve. 
    """
    userid = es.getuserid()
    if not userid:
        es.dbgmsg(0, "EST ERROR: Can't disolve when no players on the server")
        return False
    target = args[0]
    if target.isdigit():
        es.setentityname(int(target), "dissolver_target")
        target = "dissolver_target"
    index = es.createentity("env_entity_dissolver", "env_entity_dissolver")
    es.entitysetvalue(index, "target", target)
    es.entitysetvalue(index, "magnitude", 1)
    es.entitysetvalue(index, "dissolvetype", 2)
    es.spawnentity(index)
    es.server.es_fire(userid, "env_entity_dissolver", "Dissolve")
    
def changeName(args):   
    """
    Forces a player to change their name. This is not supported by the OB engine
    so this method is pointless.
    
    @param int args[0] userid / name / steamid of the user to change their name
    @param str args[1] name The new name of the player
    """
    es.cexec(es.getuserid(args[0]), 'name', args[1])
    
def uptime(args):
    """
    Gets the amount of seconds the server has been up for.
    
    @param str args[0] var The name of the variable to assign.
    """
    es.ServerVar(args[0]).set(time.time() - startTime)

def changeToUpper(args):
    """
    Converts a string to an upper case string.
    
    @param str args[0] var The name of the variable to assign to the variable
    @param str args[1] string The string to convert to an upper-case string
    """
    es.ServerVar(args[0]).set(args[1].upper())
    
def changeToLower(args):
    """
    Converts a string to a lower case string.
    
    @param str args[0] var The name of the variable to assign to the variable
    @param str args[1] string The string to convert to a lower-case string
    """
    es.ServerVar(args[0]).set(args[1].lower())
    
def getStringLength(args):
    """
    Returns the length of a string to a variable.
    
    @param str args[0] var The name of the variable to assign the value
    @param str args[1] string The string to retrieve the length of the characters
    """
    es.ServerVar(args[0]).set(len(args[1]))
    
def stripPlayer(args):
    """
    Removes all items / weapons from a player.
    
    @param str args[0] est filter List of EST filters to strip
    """
    for player in getUseridsFromFilters(args[0]):
        spe.giveNamedItem(player, "player_weaponstrip")
        es.server.es_fire(player, "player_weaponstrip", "Strip")
        gamethread.delayed(0, es.server.es_fire, (player, "player_weaponstrip", "Kill"))
        
def speed(args):
    """
    Alters the speed for a EST filter list
    
    @param str args[0] EST Filter List
    @param float args[1] speed New speed value
    """
    for player in getPlayersFromFilters(args[0]):
        player.speed = float(args[1])
        
def teleport(args):
    """
    Teleports a player to an x,y,z position on the map.
    
    @param str|int args[0] userid/name The identification for a player
    @param float args[1] x The x coord of the position
    @param float args[2] y The y coord of the position
    @param float args[3] z The z coord of the position
    """
    userid = es.getuserid(args[0])
    if userid:
        es.server.es_setpos(userid, float(args[1]), float(args[2]), float(args[3]))
        
def trimDecimal(args):
    """
    Trims a decimal down to a specific decimal positioning.
    
    @param str args[0] var The varibale to hold the trimmed result
    @param float args[1] number The number to trim
    @param int args[2] decimal place The number of decimals to trim to 
    """
    decimalPoints = int(args[2])
    string = args[1]
    if string.count(".") > 1:
        es.dbgmsg(0, "EST ERROR: String must contain only 1 decimal point, instead received %s" % string)
        return False
    elif string.count(".") == 0:
        string += ".0"
    try:  
        first, second = map(int, string.split("."))
    except ValueError:
        es.dbgmsg(0, "EST ERROR: String needs to be a valid float value, instead received %s" % string)
        return False
    if not decimalPoints:
        es.ServerVar(args[0]).set(first)
    else:
        second = str(second)[:decimalPoints]
        es.ServerVar(args[0]).set("%s.%s" % (first, second) )
        
def slay(args):
    """
    Slays a player and emits an explosion noise to everyone around them.
    
    @param str args[0] EST Player Filter
    """
    for player in getPlayersFromFilters(args[0]):
        es.emitsound('player', int(player), 'ambient/explosions/exp3.wav', '1.0', '0.4')
        es.server.damage(int(player), player.health)
        
def viewCone(args):
    """
    Retreives a list of all the players within a specified view cone, and
    executes a command string for each player within the cone.

    @param str args[0] var The variable to store the userid in
    @param str|int args[1] name|userid The identifier of the view cone start 
    @param str args[2] filter A filter for all the users to return inside the cone
    @param float args[3] coneAngle The angle of the cone (90 = player's actual view cone)
    @param float args[4] distance The distance of the view cone
    @param str args[5] command The command string to execute for each player
    """
    foundPlayers = []
    userid = es.getuserid(args[1])
    player = playerlib.getPlayer(userid)
    if not userid:
        es.dbgmsg(0, "EST ERROR: Cannot find user by ID %s" % args[1])
        return False
    originalVector = vecmath.Vector(es.getplayerlocation(userid))
    viewVector = vecmath.Vector(player.viewVector())
    for target in getUseridsFromFilters(args[2]):
        targetVector = vecmath.vector(es.getplayerlocation(target))
        vectorBetween = targetVector - originalVector
        
        try:
            angle = degrees(vecmath.angle(vectorBetween, viewVector))
        except ZeroDivisionError:
            continue

        if angle <= float(args[3]) * 0.5 and vectorBetween.length() < float(args[4]):
            es.ServerVar(args[0]).set(target)
            es.server.cmd(args[5])

def drug(args):
    """
    Drug a player which alters their vision to look distorted. This will just
    simply change their field of vision to widen the area.
    
    @param str args[0] EST Player Filter
    @param optional float args[1] time The time the effect will last (seconds)
    """
    for userid in getUseridsFromFilters(args[0]):
        es.setplayerprop(userid, 'CBasePlayer.m_iDefaultFOV', 155)
        if len(args) > 1:
            delays[userid].addDelay(float(args[1]), "undrug", _sobre, userid)
    
def _sobre(userid):
    """
    Remove the drug effect from a certain player.
    
    @param int userid The ID of the user to remove the effect
    """
    es.setplayerprop(userid, 'CBasePlayer.m_iDefaultFOV', 90)
   
def drunk(args):
    """
    Starts a drunk affect. This will constantly alter their FOV setting which
    will make their vision distorted and "zomming in / out".
    
    @param str args[0] filters EST Player filter
    @param optional float args[1] time The time the effect will last (seconds)
    """
    for userid in getUseridsFromFilters(args[0]):
        if len(args) > 1:
            _drunkLoop(userid, time.time() + float(args[1]))
        else:
            _drunkLoop(userid) 
            
def _drunkLoop(userid, endTime=None, boolDown=True):
    """
    This is the command which affects the player's FOV; essentially
    distorting their vision. This will be executed until either the delay
    is canceled or the ending time is reached. Essentially all this does is
    alter their FOV by 10 each iteration.

    @param int userid The ID of the user who this effect is for
    @param float endTime The seconds since epoch that the effect will end
    @param bool boolDown Whether or not the current effect is deducting or adding to the FOV.
    """
    if endTime is None or time.time() <= endTime:
        fov = es.getplayerprop(userid, 'CBasePlayer.m_iDefaultFOV')
        if 40 < fov < 90:
            if boolDown:
                fov -= 10
            else:
                fov += 10
        elif fov == 40:
            fov += 10
            boolDown = False
        elif fov == 90:
            boolDown = True
            fov -= 10
        es.setplayerprop(userid, 'CBasePlayer.m_iDefaultFOV', fov)
        delays[userid].addDelay(0.2, "drunk", _drunkLoop, userid, endTime, boolDown)
    else:
        es.setplayerprop(userid, 'CBasePlayer.m_iDefaultFOV', 90)

def _getWeapon(target):
    """
    Gets all the weapons objects for any weapon slots / weapon names provided.
    
    @param str|int target The weapon slot or weapon name to get the weapon objects
    """
    weapons = []
    if target.isdigit():
        for weaponObj in weaponlib.getWeaponList():
            if weaponObj.slot == target:
                weapons.append(weaponObj)
    else:
        weapon = weaponlib.getWeapon(target)
        if weapon is not None:
            weapons.append(weapon)
    return weapons
        
def getAmmo(args):
    """
    Retireves the current ammo count for a player's weapon slot / name 
    
    @param str args[0] var The name of the variable to store the return value
    @param int|str args[1] username/userid The identifier for the user
    @param str|int args[2] weapon name/slot The slot or name of the weapon to get
    """
    weapons = _getWeapon(args[2])
    
    if not weapons:
        es.dbgmsg(0, "EST ERROR: Cannot find weapon '%s' for getammo" % target)
        es.ServerVar(args[0]).set(0)
        return False
    userid = es.getuserid(args[1])
    if userid is None:
        es.dbgmsg(0, "EST ERROR: Cannot get ID of user '%s'" % args[1])
        es.ServerVar(args[0]).set(0) 
        return False
    handle = es.getplayerhandle(userid)
    ammo = 0
    for weapon in weapons:
        if ammo:           
            break
        for index in weapon.indexlist:
            if es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity') == handle:
                ammo = es.getplayerprop(userid, weapon.prop)
                break
    es.ServerVar(args[0]).set(ammo)
    
def getClip(args):
    """
    Retireves the current clip count for a player's weapon slot / name 
    
    @param str args[0] var The name of the variable to store the return value
    @param int|str args[1] username/userid The identifier for the user
    @param str|int args[2] weapon name/slot The slot or name of the weapon to get
    """
    weapons = _getWeapon(args[2])
    
    if weapon is None:
        es.dbgmsg(0, "EST ERROR: Cannot find weapon '%s' for getammo" % target)
        es.ServerVar(args[0]).set(0)
        return False
    userid = es.getuserid(args[1])
    if userid is None:
        es.dbgmsg(0, "EST ERROR: Cannot get ID of user '%s'" % args[1])
        es.ServerVar(args[0]).set(0) 
        return False
    handle = es.getplayerhandle(userid)
    ammo = 0
    for weapon in weapons:
        if ammo:           
            break
        for index in weapon.indexlist:
            if es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity') == handle:
                ammo = es.getindexprop(index, "CBaseCombatWeapon.LocalWeaponData.m_iClip1")
                break
    es.ServerVar(args[0]).set(ammo)
    
def getWeaponIndex(args):
    """
    Gets a weapon index for a player's weapon.
    
    @param str args[0] var The name of the variable to store the result
    @param str|int args[1] username/userid The Identifier of the user
    @param str|int args[2] Weapon slot / name
    """
    weapons = _getWeapon(args[2])
    
    if weapon is None:
        es.dbgmsg(0, "EST ERROR: Cannot find weapon '%s' for getammo" % target)
        es.ServerVar(args[0]).set(-1)
        return False
    userid = es.getuserid(args[1])
    if userid is None:
        es.dbgmsg(0, "EST ERROR: Cannot get ID of user '%s'" % args[1])
        es.ServerVar(args[0]).set(-1) 
        return False
    handle = es.getplayerhandle(userid)
    realIndex = -1
    for weapon in weapons:
        if realIndex > -1:           
            break
        for index in weapon.indexlist:
            if es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity') == handle:
                realIndex = index
                break
    es.ServerVar(args[0]).set(realIndex)
    
def give(args):
    """
    Gives a certain player a specific entity
    
    @param str args[0] EST Player Filter
    @param str args[1] entity The entity to give to the players
    """
    for userid in getUseridsFromFilters(args[0]):
        spe.giveNamedItem(userid, args[1])
        
def god(args):
    """
    Assigns a player to be a god (immune from all damage)
    
    @param str args[0] EST Player Filter
    @param int args[1] 1/0 Whether or not to enable (1) or disable (0) godmode. 
    """
    for player in getPlayersFromFilters(args[0]):
        player.godmode(int(args[1]))
        
def getWallBetween(args):
    """
    This gets any obstruction between two sets of coordinates. This is kinda 
    hackish. The basic aim of this is to take a random player, teleport them to
    the first set of coords, force them to "look" at the new location, then 
    create an object at their viewcoord. If that object has the same x,y,z as
    the target x,y,z then there is no obstruction. If there is a change in the
    positions, then there is an obstruction (the object is created much closer).
    After this has finished, we move the player back to the original position
    and set their view back to normal. This should occur very quickly (even before
    a frame occurs) so hopefully the user won't notice :D.
    
    @param str args[0] var The name of the variable to assignt he result to
    @param float args[1] x The x of the initial set of coords
    @param float args[2] y The y of the initial set of coords
    @param float args[3] z The z of the initial set of coords
    @param float args[4] x2 The x coord of the second set of coords
    @param float args[5] y2 The y coord of the second set of coords
    @param float args[6] z2 The z coord of the second set of coords
    """
    userid = es.getuserid()
    if userid is None:
        es.dbgmsg(0, "EST ERROR: est_GetWallBetween requires a player on a server")
        es.ServerVar(args[0]).set(0)
        return False
    x, y, z = es.getplayerlocation(userid)
    target_x, target_y, target_z = [args[i] for i in xrange(1, 4)]
    es.server.es_setpos(userid, target_x, target_y, target_z)
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(int(player.isObstructionBetween([args[i] for i in xrange(4, 7)])))
    es.server.es_setpos(userid, x, y, z)
    
def getViewAngle(args):
    """
    Retrievs the user's current pitch / yaw / roll of their view.
    
    @param str|int args[0] username/userid The ID of the user
    @param str args[1] pitchvar The name of the variable to store the pitch value
    @param str args[2] yawvar The name of the variable to store the yaw value
    @param str args[3] rollvar The name of the variable to store the roll value
    """
    userid = es.getuserid(args[0])
    if not es.exists("userid", userid):
        es.dbgmsg(0, "EST ERROR: est_GetViewAngles cannot find user %s" % args[3])
        es.ServerVar(args[1]).set(0.0)
        es.ServerVar(args[2]).set(0.0)
        es.ServerVar(args[3]).set(0.0)
        return False
    for x, y in enumerate(playerlib.getPlayer(userid).getViewAngle()):
        es.ServerVar(args[x+1]).set(y)
        
def getViewCoord(args):
    """
    Gets the player's coordinates that they are currently looking at as an x, y
    and z variable and store them in variables.
    
    @param str|int args[0] username/userid The ID of the user
    @param str args[1] x The name of the variable to store the x value
    @param str args[2] y The name of the variable to store the y value
    @param str args[3] z The name of the variable to store the z value
    """
    userid = es.getuserid(args[0])
    if not es.exists("userid", userid):
        es.dbgmsg(0, "EST ERROR: est_GetViewCoord cannot find user %s" % args[3])
        es.ServerVar(args[1]).set(0.0)
        es.ServerVar(args[2]).set(0.0)
        es.ServerVar(args[3]).set(0.0)
        return False
    if not int(es.ServerVar("sv_cheats")):
        es.forcevalue("sv_cheats", 1)
        es.sexec(userid, "prop_dynamic_create", "props_c17/tv_monitor01_screen.mdl")
        es.sexec(userid, "ent_setname viewed_prop")
        es.forcevalue("sv_cheats", 0)
    else:
        es.sexec(userid, "prop_dynamic_create", "props_c17/tv_monitor01_screen.mdl")
        es.sexec(userid, "ent_setname viewed_prop")
    propIndex = 0
    for index in es.createentitylist("prop_dynamic"):
        if (es.entitygetvalue(index, "targetname") == "viewed_prop"):
            propIndex = index
            es.entitysetvalue(index, "targetname", "")
            spe.removeEntityByIndex(index)
            break
    values = (0, 0, 0)
    if propIndex > 0:
        values = map(float, es.entitygetvalue(index, "origin").split(" "))
    for x, y in enumerate(values):
        es.ServerVar(args[x+1]).set(y)
        
def getViewPlayer(args):
    """
    Returns the player that this user is looking at. If the player is not
    looking at any other character, then state that.

    @param int|str args[0] username/userid The ID of the player to get who they are looking at
    @param str args[1] var The name of the varibale to store the result
    """
    userid = es.getuserid(args[0])
    if not es.exists("userid", userid):
        es.dbgmsg(0, "EST ERROR: Cannot find user matching %s for est_GetViewPlayer" % args[1])
        es.ServerVar(args[1]).set(0)
        return False
    foundPlayer = 0
    index = getViewProp(userid)
    if index:
        for player in playerlib.getPlayerList():
            if player.index == index:
                foundPlayer = int(player)
                break
    es.ServerVar(args[1]).set(foundPlayer) 

def getViewProp(args):
    """
    Returns the prop that this user is looking at. If the player is not
    looking at any other prop, then state that.

    @param int|str args[0] username/userid The ID of the player to get who they are looking at
    @param str args[1] var The name of the varibale to store the result
    """
    userid = es.getuserid(args[0])
    if not es.exists("userid", userid):
        es.dbgmsg(0, "EST ERROR: Cannot find user matching %s for est_GetViewProp" % args[1])
        es.ServerVar(args[1]).set(-1)
        return False
    foundIndex = 0
    if not int(es.ServerVar("sv_cheats")):
        es.forcevalue("sv_cheats", 1)
        es.sexec(userid, "ent_setname viewed_prop")
        es.forcevalue("sv_cheats", 0)
    else:
        es.sexec(userid, "ent_setname viewed_prop")
    for index in es.createentitylist():
        if (es.entitygetvalue(index, "targetname") != "viewed_prop"):
            continue
        foundIndex = index
        es.entitysetvalue(index, "targetname", "")
        break
    es.ServerVar(args[1]).set(foundIndex)
    
def getAngleToPoint(args):
    """
    Retrives the pitch, yaw and roll required to look at a specific coord on
    the map. This gets the person's current ViewVector and calculates the angle
    between the that and the required point of the (x,y,z) vector.
    
    @param str args[0] identifier The user identifier
    @param float args[1] x The x coord 
    @param float args[2] y The y coord 
    @param float args[3] z The z coord 
    @param str args[4] pitch var The variable to store the return pitch value
    @param str args[5] pitch var The variable to store the return pitch value
    @param str args[6] pitch var The variable to store the return pitch value
    """
    userid = es.getuserid(args[0])
    if not es.exists("userid", userid):
        es.dbgmsg(0, "EST ERROR: est_GetAngleToPoint cannot find user %s" % args[0])
        es.ServerVar(args[4]).set(0.0)
        es.ServerVar(args[5]).set(0.0)
        es.ServerVar(args[6]).set(0.0)
        return False
    targetVector = vecmath.Vector(args[1:4])
    viewVector = vecmath.Vector(playerlib.getPlayer(userid).viewVector())
    for x, y in enumerate(vecmath.viewangles(targetVector, viewVector)):
        es.ServerVar(args[x+4]).set(y)
        
def getChat(args):
    """
    Retrievs the latest chat message said by the specified user
    
    @param str args[0] var The name of the variable to store the last chat string
    @param str|int args[1] identifier The ID of the user to acuqire the latest chat string
    """
    chatString = ""
    userid = es.getuserid(args[1])
    if es.exists("userid", userid):
        if userid in lastMessages:
            chatString = lastMessages[userid]
    else:
        es.dbgmsg(0, "EST ERROR: Cannot find userid by given identifier '%s'" % args[1])
        es.ServerVar(args[0]).set("")
        return False
    es.ServerVar(args[0]).set(chatString)
    
def getEntityClassName(args):
    """
    Retrieves the class name of an entity index.
    
    @param str args[0] var The name of the variable to store the return value
    @param int args[1] ent index The entity index to retrieve the classname
    """
    entList = es.createentitylist()
    index = int(args[1])
    className = ""
    if index in entList:
        className = entList[index]["classname"]
        es.ServerVar(args[0]).set(className)
    else:
        es.dbgmsg(0, "EST ERROR: Cannot find entity %s" % index)
        es.ServerVar(args[0]).set("")    
        return False
        
def getEntityIndex(args):
    """
    Retrieves the first index found by classname which is owned by a player. Will
    return -1 if no index is found.
    
    @param str args[0] var The name of the variable to hold the result
    @param int|str args[1] identifier Player identifier
    @param str args[2] classname The entity classname
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetEntityIndex cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set(-1)
        return False
    index = -1
    handle = es.getplayerhandle(userid)
    for entity in es.createentitylist(args[2]):
        if es.getindexprop(entity, "CBaseEntity.m_hOwnerEntity") == handle:
            index = entity
            break
    es.ServerVar(args[0]).set(index)
    
def getEntName(args):
    """
    Gets an entity name from a given index. This will only work if the entity
    had it's name assigned by est_SetEntName.
    
    @param str args[0] var The variable to assign the name
    @param int args[1] entindex The entity index to get the name of
    """
    index = int(args[1])
    name = ""
    if index in entNames:
        name = entNames[index]
    es.ServerVar(args[0]).set(name)

def setEntName(args):
    """
    Sets an entity name for a given index. This allows us to modify it via
    es_fire etc.
    
    @param int args[0] entindex The entity index to assign a name to
    @param str args[1] name The name to assign the index.
    """
    index = int(args[0])
    name = args[1]
    es.setentityname(index, name)
    entNames[index] = name
    
def getEyeCoord(args):
    """
    Gets the eye coordinates of the player
    
    @param int|str args[0] identifier Player identifier
    @param float args[1] var_x The variable to store the X coord
    @param float args[2] var_y The variable to store the Y coord
    @param float args[3] var_z The variable to store the Z coord
    """
    userid = es.getuserid(args[0])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetEyeCoord cannot find userid %s" % args[0])
        es.ServerVar(args[1]).set(-1)
        es.ServerVar(args[2]).set(-1)
        es.ServerVar(args[3]).set(-1)
        return False
    player = playerlib.getPlayer(userid)
    for x, y in enumerate(player.getEyeCoord()):
        es.ServerVar(arg[x+1]).set(y)
        
def getGunCoord(args):
    """
    Gets the gun coordinates of the player
    
    @param int|str args[0] identifier Player identifier
    @param float args[1] var_x The variable to store the X coord
    @param float args[2] var_y The variable to store the Y coord
    @param float args[3] var_z The variable to store the Z coord
    """
    userid = es.getuserid(args[0])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetGunCoord cannot find userid %s" % args[0])
        es.ServerVar(args[1]).set(-1)
        es.ServerVar(args[2]).set(-1)
        es.ServerVar(args[3]).set(-1)
        return False
    player = playerlib.getPlayer(userid)
    for x, y in enumerate(player.getEyeCoord()):
        if x == 2:
            y -= 20
        es.ServerVar(arg[x+1]).set(y)
    
def getIP(args):
    """
    Gets the IP address of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetIp cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set("192.168.2.1")
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(str(player.address))
    
def getIndex(args):
    """
    Gets the entity index of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetIndex cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(-1)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(int(player.index))
    
def getName(args):
    """
    Gets the Name of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetName cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set("")
        return False
    es.ServerVar(args[0]).set(es.getplayername(userid))
    
def getUserid(args):
    """
    Gets the users ID of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetUserid cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(0)
        return False
    es.ServerVar(args[0]).set(userid)  
    
def getSteam(args):
    """
    Gets the SteamID of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetSteam cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(0)
        return False
    es.ServerVar(args[0]).set(es.getplayersteamid(userid))
    
def getTeam(args):
    """
    Gets the team number of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetTeam cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(0)
        return False
    es.ServerVar(args[0]).set(es.getplayerteam(userid)) 
    
def getHE(args):
    """
    Gets the grenades of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetHE cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(player.getHE())    
    
def getFB(args):
    """
    Gets the flash bangs of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetFB cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(player.getFB())     
    
def getSG(args):
    """
    Gets the smoke grenades of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetSG cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(player.getSG())    
    
def getPrimary(args):
    """
    Gets the smoke grenades of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetPrimary cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(player.getPrimary())    
    
def getSecondary(args):
    """
    Gets the smoke grenades of a player and stores the result in a variable
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetSecondary cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(player.getSecondary())   
    
def getGun(args):
    """
    Gets the player's active gun
    
    @param str args[0] var The variable to hold the result
    @param str|int args[1] ident The player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetGun cannot find userid %s" % args[0])
        es.ServerVar(args[0]).set("")
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(str(player.weapon))  
    
def getPlayerDistance(args):
    """
    Gets the distance, azimuth and elevation of 1 player from another. Distance
    is the length between the players in game units, the azimuth is the yaw
    rotation they are away from their target (view vector to players coords) and
    elevation is the pitch from eye coord to eye coord.
    
    @param str args[0] distance The variable to store the distance result
    @param str args[1] azimuth The variable to store the distance result
    @param str args[2] elevation The variable to store the distance result
    @param int|str args[3] identifier Player identifier
    @param int|str args[4] identifier Target player identifier
    """
    userid = es.getuserid(args[3])
    target = es.getuserid(args[4])
    if not es.exist('userid', userid) or not es.exists('userid', target):
        es.dbgmsg(0, "est_GetPlayerLocation cannot find userid %s or target %s" % (args[3], args[4]))
        es.ServerVar(args[0]).set(-1)
        es.ServerVar(args[1]).set(-1)
        es.ServerVar(args[2]).set(-1)
        return False
    player = playerlib.getPlayer(userid)
    targetPlayer = playerlib.getPlayer(target)
    startVector = vecmath.Vector(player.getEyeCoord())
    targetVector = vecmath.Vector(targetPlayer.getEyeCoord())
    viewVector = vecmath.Vector(player.viewVector())
    distance = (targetVector - startVector).length()
    pitch, yaw, roll = vecmath.viewangles(targetVector, viewVector)
    es.ServerVar(args[0]).set(distance)
    es.ServerVar(args[1]).set(yaw)
    es.ServerVar(args[2]).set(pitch)
    
def getSlot(args):
    """
    Returns the weapon name for a player's weapon at the designated slot. e.g.
    if the slot is 3 it will return weapon_knife.
    
    @param str args[0] var The name of the variable to store the result
    @param int|str args[1] identifier Player identifier (either name or userid)
    @param int args[2] slot The slot of the weapon
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetSlot cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set("")
        return False
    weapons = _getWeapon(args[2])
    if weapon is None:
        es.dbgmsg(0, "Weapon slot %s is not a valid weapon slot" % args[2])
        es.ServerVar(args[0]).set("")
        return False
    handle = es.getplayerhandle(userid)
    weaponName = ""
    for weapon in weapons:
        if weaponName:
            break
        for index in weapon.indexlist:
            if es.getindexprop(index, "CBaseEntity.m_hOwnerEntity") == handle:
                weaponName = weapon.name
                break
    es.ServerVar(args[0]).set(weaponName)
    
def getHelp(args):
    """
    Prints the debug message for a command.
    
    @param optional args[0] command The name of the command to get help.
    """
    if args:
        est[args[0]].displayHelp()
    else:
        est.dumpCommands()
        
def setGravity(args):
    """
    Sets the gravity for an EST player filter to a specified ratio with regards
    to the server's current gravity.
    
    @param str args[0] EST Player filter
    @param float args[1] ratio The ratio to the current gravity (1.0 = normal)
    """
    amount = float(args[1])
    for userid in getUseridsFromFilters(args[0]):
        if userid not in gravity:
            if amount != 1.0:
                gravity.addPlayer(userid, amount)
        else:
            if amount == 1.0:   
                del gravity[userid]
            else:
                gravity[userid].amount = amount
    
def getGravity(args):
    """
    Retrieves the current gravity value for a specific player.
    
    @param str args[0] var The name of the variable to hold the result
    @param int|str args[1] identifier Player identifier (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_GetGravity cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set(0)
        return False
    amount = 1.0
    if userid in gravity:
        amount = gravity[userid].amount
    es.ServerVar(args[0]).set(amount)
    
def fade(args):
    """
    Apply on overlay to a person's screen and starts a loop which will
    continue update the blind throughout the delay time.

    @param str args[0] EST Player Filter
    @param int args[1] type The type of the fade:
                      - 0 = No fade
                      - 1 = Fade Out
                      - 2 = fade in
    @param float args[2] fade time The amount of time it takes to fade the screen (unused)
    @param float args[3] total time The amount of time that the person will be blinded 
    @param int args[4] red The red pigment of the overlay
    @param int args[5] green The green pigment of the overlay
    @param int args[6] blue The blue pigment of the overlay
    @param int args[7] alpha The opacity of the overlay
    """
    delay = float(args[3])
    red, green, blue, alpha = (map(int, (args[x] for x in xrange(4, 8))))
    type = int(args[1])
    for userid in getUseridsFromFilters(args[0]):
        delays[userid].cancelDelay("blind")
        __updateFade(userid, time.time() + delay, alpha, type, red, green, blue)

def __updateFade(userid, endTime, alpha, type = 2, red = 0, green = 0, blue = 0):
    """
    Since the usermessage fade works on FPS, we need to keep reapplying
    the fade so we can cancel it whenever we want.
    
    @internal
    @param int userid The ID of the user this effects
    @param float endTime The system's time since epoch that the fade ends
    @param int alpha The opacity of the blind
    @param int type The type of the blind
    @param int red The red pigment of the overlay
    @param int green The green pigment of the overlay
    @param int blue The blue pigment of the overlay
    """
    if time.time() < endTime:
        es.usermsg('create', 'fade', 'Fade')
        es.usermsg('write', 'short', 'fade', 5)     # Frames To fade
        es.usermsg('write', 'short', 'fade', 45)    # Frames to stay faded
        es.usermsg('write', 'short', 'fade', type)  # Type of fade
        es.usermsg('write', 'byte',  'fade', red)   # Red
        es.usermsg('write', 'byte',  'fade', green) # Green
        es.usermsg('write', 'byte',  'fade', blue)  # Blue
        es.usermsg('write', 'byte',  'fade', alpha) # Alpha
        es.usermsg('send', 'fade', userid)
        es.usermsg('delete', 'fade')
        delays[userid].addDelay(0.1, "blind", __updateFade, userid, endTime, alpha, type, red, green, blue)
        
def hasDefuser(args):
    """
    Returns whether or not a player has a defusal kit on them.
    
    @param str args[0] var The name of the variable to store the result
    @param str|int args[1] ident The ID of the user (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_HasDefuser cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(int(player.hasDefuser()))
    
def hasHelmet(args):
    """
    Returns whether or not a player has a kevlar helmet on them.
    
    @param str args[0] var The name of the variable to store the result
    @param str|int args[1] ident The ID of the user (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_HasHelmet cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(int(player.hasHelmet()))
    
def hasKevlar(args):
    """
    Returns whether or not a player has a kevlar on them.
    
    @param str args[0] var The name of the variable to store the result
    @param str|int args[1] ident The ID of the user (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_HasKevlar cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(int(bool(player.getArmor())))
    
def isAlive(args):
    """
    Returns whether or not a player is alive.
    
    @param str args[0] var The name of the variable to store the result
    @param str|int args[1] ident The ID of the user (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_IsAlive cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set(0)
        return False
    es.ServerVar(args[0]).set(es.getplayerprop(userid, 'CBasePlayer.pl.deadflag'))
    
def isBomber(args):
    """
    Returns whether or not a player has the bomb
    
    @param str args[0] var The name of the variable to store the result
    @param str|int args[1] ident The ID of the user (username / userid)
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_IsBomber cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set(0)
        return False
    player = playerlib.getPlayer(userid)
    es.ServerVar(args[0]).set(int(player.hasC4()))
    
def isAlpha(args):
    """
    Returns whether or string contains just alpha characters (a-zA-Z)
    
    @param str args[0] var The name of the variable to store the result
    @param str args[1] string The string to test for validation
    """
    es.ServerVar(args[0]).set(int(args[1].isalpha()))
    
def isOnly(args):
    """
    Returns whether or string contains just the characters specified
    
    @param str args[0] var The name of the variable to store the result
    @param str args[1] string The string to test for validation
    @param str args[2] allowed chars A string of allowed characters
    """
    testString = args[1]
    for char in args[2]:
        testString = testString.replace(char, "")
    es.ServerVar(args[0]).set(int(not bool(testString)))
    
def isNumeric(args):
    """
    Returns whether or string contains just numeric characters (0-9)
    
    @param str args[0] var The name of the variable to store the result
    @param str args[1] string The string to test for validation
    """
    try:
        int(args[1])
        es.ServerVar(args[0]).set(1)
    except ValueError:
        es.ServerVar(args[0]).set(0)
        
def jetpack(args):
    """
    Sets jetpack on a lits of players.
    
    @param str args[0] EST Player Filter
    @param int args[1] 1/0 Enable / Disable jetpack
    """
    for player in getPlayersFromFilters(args[0]):
        player.jetpack = int(args[1])
        
def noclip(args):
    """
    Sets noclip on a lits of players.
    
    @param str args[0] EST Player Filter
    @param int args[1] 1/0 Enable / Disable noclip
    """
    for player in getPlayersFromFilters(args[0]):
        player.noclip = int(args[1])
    
def motdFromRawText(args):
    """
    Displays a MOTD window from a given URL value
    
    @param str args[0] EST player filter
    @param str args[1] title The title of the MOTD windows
    @param str args[2] command UNUSED!
    @param str args[3] message The raw text which will be displayed
    """
    usermsg.motd(getUseridsFromFilters(args[0]), 0, args[1], args[3])
    
def motdFromURL(args):
    """
    Displays a MOTD window from a given URL value
    
    @param str args[0] EST player filter
    @param str args[1] title The title of the MOTD windows
    @param str args[2] command UNUSED!
    @param str args[3] URL The URL of the page which will be displayed here
    """
    usermsg.motd(getUseridsFromFilters(args[0]), 2, args[1], args[3])
    
def motdFromFile(args): 
    """
    Displays a MOTD window from a given file value
    
    @param str args[0] EST player filter
    @param str args[1] title The title of the MOTD windows
    @param str args[2] command UNUSED!
    @param str args[3] file The path to the file of the text which will be displayed here
    """
    usermsg.motd(getUseridsFromFilters(args[0]), 3, args[1], args[3])
    
def message(args):
    """
    Sends a message to the a group of players
    
    @param str args[0] EST Player Filter
    @param str args[1] message The message to send to the players
    """
    for userid in getUseridsFromFilters(args[0]):
        es.tell(userid, '#multi', args[1])
        
def near(args):
    """
    Executes a console command for every player in range of a source userid
    
    @param str args[0] var The name of the variable which will be used to store the tempoary userid
    @param str args[1] est filter The filter that this near will pick up
    @param int args[2] distance The distance of the near command
    @param int|str args[3] identifier Player identifier to use as a source x,y,z
    @param str args[4] command The command string to execute for each player
    """
    userid = es.getuserid(args[3])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_Near cannot find userid %s" % args[3])
        es.ServerVar(args[0]).set(0)
        return False
    source_x, source_y, source_z = es.getplayerlocation(userid)
    for player in getUseridsFromFilter(args[1]):
        if player == userid:
            continue
        x, y, z = es.getplayerlocation(player)
        distance = ((source_x - x) ** 2 + (source_y - y) ** 2 + (source_z - z) ** 2) ** 0.5
        if distance <= float(args[2]):
            es.ServerVar(args[0]).set(player)
            es.server.cmd(args[4])
            
def nearCoord(args):
    """
    Executes a console command for every player in range of an x,y,z source 
    
    @param str args[0] var The name of the variable which will be used to store the tempoary userid
    @param str args[1] est filter The filter that this near will pick up
    @param int args[2] distance The distance of the near command
    @param float args[3] x Source X Location
    @param float args[4] y Source Y Location
    @param float args[5] z Source Z Location
    @param str args[6] command The command string to execute for each player
    """
    source_x, source_y, source_z = map(float, args[3:6])
    for player in getUseridsFromFilter(args[1]):
        if player == userid:
            continue
        x, y, z = es.getplayerlocation(player)
        distance = ((source_x - x) ** 2 + (source_y - y) ** 2 + (source_z - z) ** 2) ** 0.5
        if distance <= float(args[2]):
            es.ServerVar(args[0]).set(player)
            es.server.cmd(args[6])
            
def knifeOnly(args):
    """
    Returns whether or not a player only has a knife (C4 is allowed as well.)
    
    @param str args[0] var The name of the variable to store the result
    @param str|int args[1] ident The ID of the user
    """
    userid = es.getuserid(args[1])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_KnifeOnly cannot find userid %s" % args[1])
        es.ServerVar(args[0]).set(0)
        return False
    handle = es.getplayerhandle(userid)
    knifeOnly = 1
    for weapon in weaponlib.getWeaponList():
        if not knifeOnly:
            break
        if weapon.name == "weapon_c4" or weapon.name == "weapon_knife":
            continue
        for index in weapon.indexlist:
            if es.getplayerhandle(userid, 'CBaseEntity.m_hOwnerEntity') == handle:
                knifeOnly = 0
                break
    es.ServerVar(args[0]).set(knifeOnly)
    
def changeLightStyle(args):
    """
    Changes the current brightness and style of the map (i.e make it darker)
    
    @param str args[0] stylestring The type of the light (a-z with z being darkest)
    @param int args[1] stylenum The number of the lightstyle (0 - 63, usually 0)
    """
    es.lightstyle(args[1], int(args[0]))
    
def makeEntity(args):
    """
    Creates an entity at a given x,y,z position with the model passed. Also
    optionally returning the index into a return value.
    
    @param str args[0] entity The type of entity to create
    @param str args[1] model The name of the model to create
    @param float args[2] x The x coordinate that the model will spawn
    @param float args[3] y The y coordinate that the model will spawn
    @param float args[4] z The z coordinate that the model will spawn
    @param optional str args[5] var The name of the variable to hold the index
    """
    index = es.createentity(args[0])
    myModel = args[1].replace("\\", "/")
    if not myModel.startswith("models/"):
        myModel = "models/" + myModel
    if not myModel.endswith(".mdl"):
        myModel = myModel + ".mdl"
    es.entitysetvalue(index, "origin", " ".join(args[2:5]))
    es.entitysetvalue(index, "World Model", myModel)
    es.spawnentity(index)
    if len(args) == 6:
        es.ServerVar(args[5]).set(index)
        
def mutePlayer(args):
    """
    Mutes a group of players. This only mutes their chat, not their voice.
    
    @param str args[0] EST Player filter
    """
    for userid in getUseridsFromFilters(args[0]):
        muted.add(userid)
        
def unmutePlayer(args):
    """
    Unmutes a group of players. This only affects their chat, not voices.
    
    @param str args[0] EST Player fitler
    """
    for userid in getUseridsFromFilters(args[0]):
        if userid in muted:
            muted.remove(userid)
            
def switchTeam(args):
    """
    Changes a player's team without killing them.
    
    @param str args[0] EST Filter
    @param int args[1] team The new team index
    """
    team = int(args[1])
    if not 0 <= team <= 3:
        es.dbgmsg(0, "est_Team requires that you specify an index between 0 and 4, you gave %s" % team)
        return False
    for userid in getUseridsFromFilters(args[0]):
        spe.switchTeam(userid, team)
        
def spawn(args):
    """
    Respawns a dead player. If they are already alive and force is enabled, then
    it will send them back to spawn and reset their health etc.
    
    @param str args[0] EST Player filter
    @param int args[1] optional force If enabled and the player is alive, they will still be respawned
    """
    forceSpawn = False
    if len(args) == 2:
        if args[1] != "0":
            forceSpawn = True
    for player in getUseridsFromFilters(args[0]):
        if es.getplayerprop(player, "CBasePlayer.pl.deadflag") or forceSpawn:
            spe.respawn(player)

def physPush(args):
    """
    Executed a physical push on a player entity (i.e alters their velocity)
    
    @param str args[0] EST Player Filter
    @param float args[1] x Physical X push
    @param float args[2] y Physical Y push
    @param float args[3] z Physical Z push 
    """
    for player in getPlayersFromFilters(args[0]):
        player.push(*map(float, args[1:]))
        
def pushTo(args):
    """
    Pushes a player to a coordinate on the map with a given force. This will
    most likely not reach their destination as objects such as walls will block
    the direct path or slow their velocity down.
    
    @param str args[0] EST Player FIlter
    @param float args[1] x The x coord on the map
    @param float args[2] y The y coord on the map
    @param float args[3] z The z coord on the map
    @force float args[4] force The force at which to push them
    """
    for userid in getUseridsFromFilters(args[0]):
        startingVector = vecmath.Vector(es.getplayerlocation(userid))
        endingVector = vecmath.Vector((x, y, z))
        vectorBetween = endingVector - startingVector
        vectorBetween *= force
        es.setplayerprop(userid, "CBasePlayer.localdata.m_vecBaseVelocity", str(vectorBetween))
        
def remove(args):
    """
    Removes an entity by a given index. We will check if the entity is within
    normal range, then check to make sure that the entity actually exists.
    
    @param int args[0] entindex The entity index to remove
    """
    index = int(args[0])
    if not spe.removeEntityByIndex(index):
        es.dbgmsg(0, "EST ERROR: Cannot remove entity index %s" % index)
        
def roundDecimal(args):
    """
    Rounds a decimal to a given decimal positioning.
    
    @param str args[0] var The name of the vairbale to hold the result
    @param float args[1] number The original number to round
    @param int args[2] decimal points The decimal positioning to round to
    """
    number = float(args[1])
    number = round(number, int(args[2]))
    es.ServerVar(args[0]).set(number)

def rocket(args):
    """
    Creates a rocket effect and pushes the player up until they explode. This is
    a better slay effect but with a delayed response.
    
    @params str args[0] EST Player filter
    """
    for userid in getUseridsFromFilters(args[0]):    
        es.emitsound('player', userid, 'weapons/rpg/rocketfire1.wav', '1.0', '0.4')
        
        es.usermsg('create', 'shake', 'Shake')
        es.usermsg('write', 'byte',  'shake', 0)
        es.usermsg('write', 'float', 'shake', 25)
        es.usermsg('write', 'float', 'shake', 1.0)
        es.usermsg('write', 'float', 'shake', 3)
        es.usermsg('send', 'shake', userid)
        es.usermsg('delete', 'shake')
        
        delays[userid].addDelay(0.5, "rocket", es.emitsound, 'player', userid, 'ambient/fire/ignite.wav', '1.0', '0.4')
        rocketEffectLoop(userid, 3.0)
    
def rocketEffectLoop(userid, time):
    """
    A command to make some neat effects and push a player vertically. This
    will kill the player if the time has ran out.
    
    @param str|int userid The ID of the user
    @param float time The time left until this effect runs out and kills the player.
    """
    if time >= 0:
        time -= 0.1
        loc = es.getplayerlocation(userid)
        if time < 2.5:
            effectlib.drawBox(loc, [loc[0] + 10, loc[1] + 10, loc[2]], "materials/sprites/laser.vmt", "materials/sprites/halo01.vmt", 0.2, 20, 20, 255, 0, 0, 255, 10, 0, 0, 0, 0)
            effectlib.drawBox(loc, [loc[0] - 10, loc[1] - 10, loc[2]], "materials/sprites/laser.vmt", "materials/sprites/halo01.vmt", 0.2, 20, 20, 255, 128, 0, 255, 10, 0, 0, 0, 0)
            effectlib.drawBox([loc[0] - 5, loc[1] - 5, loc[2]], [loc[0] + 5, loc[1] + 5, loc[2]], "materials/sprites/laser.vmt", "materials/sprites/halo01.vmt", 0.2, 20, 20, 0, 0, 255, 255, 10, 0, 0, 0, 0)
            if (time * 10) % 2:
                if es.exists('userid', userid):
                    playerlib.getPlayer(userid).set("push", (0, 145, True) )
                else:
                    return 
        delays[userid].addDelay(0.1, "rocket", rocketEffectLoop, userid, time)
    else:
        spe.giveNamedItem(userid, "env_explosion")
        es.server.es_fire(userid, "env_explosion", 'addoutput "iMagnitude 100"')
        es.server.es_fire(userid, "env_explosion", "explode")
        es.server.damage(userid, es.getplayerprop(userid, 'CBasePlayer.m_iHealth'))
        es.emitsound('player', userid, 'ambient/explosions/exp3.wav', '1.0', '0.4')
        
def playSoundForPlayers(args):
    """
    Plays a sound for individual players. Only the players passed into the 
    player filter will hear the sound. This will not emit a sound.
    
    @param str args[0] EST Player Filter
    @param str args[1] sound Path to sound file.
    """
    for userid in getUseridsFromFilters(args[0]):
        es.playsound(userid, args[1], 1.0)
        
def emitSoundFromEntity(args):
    """
    Emits a sound from a specific entity. This makes the sound audible to
    players within the map.
    
    @param int args[0] entindex The entity index the sound will be emitted from
    @param str args[1] sound Path to sound file
    @param float optional args[2] volume How loud the sound is
    @param str unused optional args[3] filters Who hears the sound (only supports #a) 
    """
    volume = 1.0
    if len(args) >= 3:
        volume = float(args[2])
    es.emitsound("entity", int(args[0]), str(args[1]), volume, 0.7)
    
def emitSoundFromPlayer(args):
    """
    Emits a sound from a specific player. This makes the sound audible to
    players within the map.
    
    @param int args[0] userid The user the sound will be emitted from
    @param str args[1] sound Path to sound file
    @param float optional args[2] volume How loud the sound is
    @param str unused optional args[3] filters Who hears the sound (only supports #a) 
    """
    userid = es.getuserid(args[0])
    if not es.exist('userid', userid):
        es.dbgmsg(0, "est_PlayPlayer cannot find userid %s" % args[0])
        return False
    volume = 1.0
    if len(args) >= 3:
        volume = float(args[2])
    es.emitsound("player", userid, str(args[1]), volume, 0.7)
    
def getPlayerCount(args):
    """
    Returns the amount of players in the given filter list.
    
    @param str args[0] var The name of the variable to hold the result
    @param str args[1] EST player filter
    """
    es.ServerVar(args[0]).set(len(getUseridsFromFilters(args[1])))
    
def getPropCount(args):
    """
    Returns the amount of entities on the current map (any more than 1024 usually
    crashes the server.)
    
    @param str args[0] var The name of the variable to hold the result
    """
    es.ServerVar(args[0]).set(len(es.createentitylist()))
    
def listAllProps(args):
    """
    Debugs a list of all the props to console and their entity ID.
    """
    entList = es.createentitylist()
    for index in entList:
        es.dbgmsg(0, "%s - %s" % (index, entList[i]["classname"]))
        
def showAllOffsets(args):
    """
    Dumps all the server offsets for the relevant entities on the current
    map.
    
    @param str args[0] unused class type Not used...
    """
    es.server.es_dumpserverclasses()
    
def getRandomPlayer(args):
    """
    Returns a random player from an EST player filter.
    
    @param str args[0] var The name of the variable used to store the random ID
    @param str args[0] EST Player Filter
    """
    filterList = getUseridsFromFilters(args[1])
    userid = 0
    if filterList:
        userid = random.choice(filterList)
    es.ServerVar(args[0]).set(userid)
    
def removeIdle(args):
    """
    Remove all entities under the classname who have no owner (handle of -1)
    
    @param optional str args[0] classname match
    """
    matchClassName = None
    if args:
        matchClassName = args[0]
    entList = es.createentitylist()
    for i in entList:
        if matchClassName is not None:
            if not entList[i]["classname"].startswith(matchClassName):
                continue
        if es.getindexprop(i, "CBaseEntity.m_hOwnerEntity") == -1:
            spe.removeEntityByIndex(i)
            
def removeWeapon(args):
    """
    Removes a weapon for a given player.
    
    @param str args[0] EST Player Filter
    @param str args[1] Weapon name / slot
    """
    weapons = _getWeapon(args[1])
    for userid in getUseridsFromFilters(args[0]):
        for weapon in weapons:
            index = spe.getWeaponIndex(userid, weapon.name)
            if index is not None:
                spe.removeEntityByIndex(index)
                es.cexec(userid, "lastinv")
            
def setEntModel(args):
    """
    Assign a different model to an entity ID.
    
    @param int args[0] index The index that the model will be altered
    @param str args[1] path The path of the model
    """
    index = int(args[0])
    previousColor = es.getindexprop(index, "CBaseEntity.m_clrRender")
    r, g, b, a = tuple(int(x) for x in (color & 0xff, (color & 0xff00) >> 8, (color & 0xff0000) >> 16, (color & 0xff000000) >> 24))
    es.setindexprop(index, "CBaseEntity.m_nModelIndex", es.precachemodel(args[1]))
    setEntColor((index, r, g, b, a))
    
def setEntColor(args):
    """
    Reassigns an entities color.
    
    @param int args[0] index The index that the color will be altered
    @param int args[1] red The red pigment (0 - 255)
    @param int args[2] green The green pigment (0 - 255)
    @param int args[3] blue The blue pigment (0 - 255)
    @param int args[4] alpha The alpha pigment (0 - 255)
    """
    index, r, g, b = map(int, args[:4])
    alpha = es.getindexprop(index, "CBaseEntity.m_nRenderMode") & 0xff000000 >> 24
    if len(args) == 5:
        alpha = int(args[4])
    color = r + (g << 8) + (b << 16) + (a << 24)
    if color >= 2**31: 
        color -= 2**32
    es.setindexprop(index, "CBaseEntity.m_nRenderMode", es.getindexprop(index, "CBaseEntity.m_nRenderMode") | 1)
    es.setindexprop(index, "CBaseEntity.m_nRenderFX", es.getindexprop(index, "CBaseEntity.m_nRenderFX") | 256)
    es.setindexprop(index, "CBaseEntity.m_clrRender", color)
    
def setPlayerColor(args):
    """
    Reassigns player's colours. This is a short wrapper for setEntColor as it
    will do the same job, but work on player filters and get their index
    automatically.
    
    @param str args[0] EST Player Filter The players whose colours will be altered
    @param int args[1] red The red pigment (0 - 255)
    @param int args[2] green The green pigment (0 - 255)
    @param int args[3] blue The blue pigment (0 - 255)
    @param int args[4] alpha The alpha pigment (0 - 255)
    """
    for player in getPlayersFromFilters(args[0]):
        args = [player.index] + args[1:]
        setEntColor(args)
        
        
def setPlayerModel(args):
    """
    Reassigns a player's model. Similarly to setPlayerColor, this is a simple 
    wrapper which automatically gets player's indexes.
    
    @param str args[0] EST Player Filter
    @param str args[1] path The path of the model
    """
    for player in getPlayersFromFilters(args[0]):
        args = (player.index, args[1])
        setEntModel(args)
        
def shake(args):
    """
    Shakes a player's screen for a given violence and time. This distorts the
    player's view.
    
    @param str args[0] EST Player Filter - who the shake will apply to
    @param float args[1] time The length of time the shake will last
    @param int args[2] distance How far the shake will travel
    @param int args[3] voilence The violence of the shake
    """
    es.usermsg('create', 'shake', 'Shake')
    es.usermsg('write', 'byte', 'shake', 0)
    es.usermsg('write', 'float', 'shake', float(args[3]))
    es.usermsg('write', 'float', 'shake', float(args[2]))
    es.usermsg('write', 'float', 'shake', float(args[1]))
    for userid in getUseridsFromFilters(args[0]):
        es.usermsg('send', 'shake', userid)
    es.usermsg('delete', 'shake')
    
def stopSound(args):
    """
    Stops a sound started by es.playsound(). This will not stop any sound the
    engine makes.
    
    @param str args[0] EST Player Filter
    @param str args[1] sound The path of the sound to stop 
    """
    for player in getUseridsFromFilters(args[0]):
        es.stopsound(player, args[1])
        
def setAim(args):
    """
    Forces a player to look in another direction.
    
    @param str args[0] EST Player Filter
    @param str|int args[1] Who to look at
    @param optional float args[2] How much to add or deduct from the z axis
    """
    target = es.getuserid(args[1])
    if not es.exists("userid", userid):
        es.dbgmsg(0, "EST ERROR: est_SetAim cannot force the aim to user %s" % args[1])
        return False
    x, y, z = es.getplayerlocation(target)
    if len(args) == 3:
        z += float(args[2])
    for player in getPlayersFromFilters(args[0]):
        player.lookAt((x,y,z))
        
def setViewCoord(args):
    """
    Forces a group of players to look at a particular coordinate
    
    @param str args[0] EST Player Filter
    @params float args[1] x The X coord to look at
    @params float args[2] y The Y coord to look at
    @params float args[3] z The Z coord to look at
    """
    x, y, z = map(float, args[1:])
    for player in getPlayersFromFilters(args[0]):
        player.lookAt((x,y,z))
        
def setViewAngle(args):
    """
    Forces a group of players to look with a particular pitch / yaw / roll
    
    @param str args[0] EST Player Filter
    @params float args[1] pitch The pitch angle (-90 to + 90)
    @params float args[2] yaw The yaw angle (0 to 360)
    @params float args[3] roll The roll angle (0 to 360)
    """
    pitch, yaw, roll = map(float, args[1:])
    for userid in getUseridsFromFilters(args[0]):
        es.server.es_setang(userid, pitch, yaw, roll)
        
def showMenu(args):
    """
    Displays a popup on the screen. Does anyone still use this over popuplib?
    
    @param float args[0] time How long the popup will last
    @param str args[1] EST Player filter
    @param str args[2] Menu string
    """ 
    for userid in getUseridsFromFilters(args[1]):
        es.menu(float(args[0]), userid, args[2])
        
def setAmmo(args):
    """
    Sets the ammo for a given player and weapon name or slot 
    
    @param str args[0] EST Player filter
    @param str|int args[1] Weapon name / slot
    @param int args[2] amount The amount of ammo the gun will have
    """
    weapons = _getWeapon(args[1])
    ammo = int(args[2])
    for userid in getUseridsFromFilters(args[0]):
        for weapon in weapons:
            if spe.ownsWeapon(userid, weapon.name):        
                es.setplayerprop(userid, weapon.prop, ammo)
        
def setClip(args):
    """
    Sets the clip for a given player and weapon name or slot 
    
    @param str args[0] EST Player filter
    @param str|int args[1] Weapon name / slot
    @param int args[2] amount The amount of ammo the gun's clip will have
    """
    weapons = _getWeapon(args[1])
    ammo = int(args[2])
    for userid in getUseridsFromFilters(args[0]):
        for weapon in weapons:
            index = spe.getWeaponIndex(userid, weapon.name)
            if index is not None:
                es.setindexprop(index, "CBaseCombatWeapon.LocalWeaponData.m_iClip1", ammo)